CREATE OR REPLACE package body hsu_brtr
/*****************************************************************************************
Purpose         Transform Business Rule Functions to Business Rule Design
                Definitions (BRDD's)

Remarks         Note that currently the utility will only create BRDD's,
                it will not update or maintain them.

Revision History  (most recent on top!)
When          Who                          Construct
  Revision    What
------------------------------------------------------------------------------------------
25-aug-2005   Lucas Jellema (AMIS)
  6.5.3.4AMIS Added new parameter and supporting code that allows user to indicate 
              whether existing Business Rule Design Definitions may be/should be overwritten
			  Also added code that writes a list of all parameters in the PL/SQL Block
			  Also added code to create Arguments (parameters) for primary key columns
10-mar-2004   Marc Vahsen                  determine_rule_type
  6.5.3.4     Use parameter p_create_cco in determine_rule_type
25-feb-2004   Sandra Muller                install, run, transform_stage2, transform_br_fun
                                           determine_rule_type, create_brdd_or_cc
  6.5.3.3     Bugfix 2761012: Add support for creating business rules as check constraints
14-jan-2002   Sandra Muller                transform_funevt, transform_stage2
  6.5.3.2     Postpone processing of entity usages until stage 2, is more efficient
              + implemented enh.req. 2711392: if delete rule, create old args
              + corrected 2 small errors in previous restructuring
13-jan-2002   Sandra Muller                transform_br_fun
  6.5.3.1     Split up transform_br_fun in small transform_br_fun, transform_stage1,
              transform_funevt and transform_funent_as_funevt
              + bugfix in transform_funevt: also process supertype usages
06-aug-2002   Sandra Muller                package specification
  6.5.2.3     Fix for bug 2481957: default value for Oracle 9i was for wrong parameter
11-feb-2002   Michiel Jonkers              relation_as_fk, relend_owned
  6.5.2.2     Relationships without corresponding FK do not result in error but warning.
              Relationships might be implemented as domain key constraints.
24-jan-2002   Marcel Kappel
  6.5.2.1     fix for bug 1885204          transform_br_fun
              replaced: ||' Business Rule '||r_fun.fun_function_label||' has not been transformed.'
              with:     ||' This triggering event is ignored.'
16-jan-2002   Sandra Muller
  6.5.2.0     fix for Oracle 9i: added default value in package spec
12-mar-2001   Sandra Muller
  6.5.1.3     When looking for existing PLM's, look in right context container,
              + added known limitation in online help: CDA error if table not accessible
19-jan-2001   Sandra Muller
  6.5.1.2     fix for bug 1603181
10-jan-2001   Lauri Boyd
  6.5.1.1     - change help text of message prefix parameter to mention apps
03-jan-2001   Sandra Muller
  6.5.1.0     Fix for bug 1566730: warning if no triggering events for a rule
16-nov-2000   Sandra Muller
  6.5.0.3     Changed reference to RuleFrame information
26-oct-2000   MARC VAHSEN
  6.5.0.2     Procedure get_html_tagged_text moved from hsu_capi to bltext.
20-oct-2000   Marc Vahsen
  6.5.0.1     Change in parameter list of hsu_capi.find_plm
19-oct-2000   Marc Vahsen
  6.5.0.0     Upgrade to designer 6i
              + handle message prefix <default>
              + business rule naming convention via function hsu_name.is_business_rule
              + message when no rules for entities with call to hsu_name.br_notification
13-sep-2000   Sandra Muller       run
  6.0.3.2     Give user friendly error message if no business rules found for given entities
              + end processing of function immediately when no rule type can be determined
11-sep-2000   Sandra Muller       add_tbl
  6.0.3.1     When clause: use or to connect them instead of and
30-aug-2000   Sandra Muller       transform_stage2, determine_rule_type
  6.0.3.0     Add old column arguments when the rule is a Transition Rule
03-aug-2000   Sandra Muller       add_tbl, process_subtype
  6.0.2.0     Fix for bug 1371205: no multiple identical when clauses
              + no when condition for supertype entity
26-jun-2000   Sandra Muller       rel_found
  6.0.1.1.    Remove spaces from other entity name before looking in remark without spaces
26-jun-2000   Michiel Jonkers     rel_found
  6.0.1.0     Bugfix 1339265: if one entity usage remark contained two different
              relationships with same name (but to different entities), only one was found
27-apr-2000   Lauri L. Boyd       determine_rule_type
  6.0.0.6     if name is not standard, use to return error and stop.  changed so it returns
              warning and does not attempt to transform that function, but continues.
03-apr-2000   Sandra Muller       install
  6.0.0.5     changed help text of create usages parameter: even if usages already exist
30-mar-2000   Sandra Muller       check_relend_event
  6.0.0.4     look for <RL> ... </RL> tags in Event Description
29-mar-2000   Sandra Muller       transform_stage2
  6.0.0.3     Use dummy message for change event rules
23-mar-2000   Sandra Muller       transform_br_fun, transform_stage2, process_subtype
  6.0.0.2     used blfun.c_fun instead of own c_fun
              + don't process common functions
              + warning if no table for triggering entity
              + mention that no retrieve usage instead of only no usage
              + on decomposition, use table alias in name
              + subtype rules: no discr col argument (+ moved to process_subtype)
21-mar-2000   Sandra Muller       transform_stage2
  6.0.0.1     added parameter p_rule_type, before this the rule type was always static
              + add_tbl: switched then and else if p_when is null (bugfix when condition)
15-mar-2000   Sandra Muller       install + several others
  6.0.0.0     improved revision history, help text, prompts, layout etc
              + error if none of selecting parameters are entered
              + don't use user_defined property 19 for relend id
              + don't call hsu_crbr.maintain_brdd if brdd already exists
                (to prevent overwriting the validation code)
11-03-2000    Sigrid Gylseth      whole package
  1.1         Completed first version. Changes and additions all through the package.
12-01-2000    Lucas Jellema       whole package
  1.0         Initial Creation
*****************************************************************************************/
/*
$add_to_revision_history
$
*/
is

--
-- private constants
--
PACKAGE_NAME     constant   varchar2(30)   := 'HSU_BRTR';
UTILITY_NAME     constant   varchar2(60)   := 'Business Rule Design Transformer';
REVISION_LABEL   constant   varchar2(30)   := '$x.y:: 6.5.3.4AMIS  $';
ON_CONDITION_CREATE constant ci_events.on_condition%type := 'CREATE';
ON_CONDITION_UPDATE constant ci_events.on_condition%type := 'UPDATE';
ON_CONDITION_DELETE constant ci_events.on_condition%type := 'DELETE';

PRIMARY_KEY_TAG constant   varchar2(30)   := '<PK/>'; -- 6.5.3.4AMIS

-- following two define dummy message for change events:
DUMMY_MSG_PREFIX      constant varchar2(3)  := 'QMS';
DUMMY_MSG_NUMBER      constant varchar2(5)  := '00151';

type t_trg_tbl_rec is record
( tbl_id          ci_table_definitions.id%type
, ins             boolean
, upd             boolean
, del             boolean
, when_condition  varchar2(2000)
);

type t_trg_col_rec is record
( tbl_id          ci_table_definitions.id%type
, col_id          ci_columns.id%type
);

type t_trg_tbl_tbl is table of
t_trg_tbl_rec
index by binary_integer
;

type t_trg_col_tbl is table of
t_trg_col_rec
index by binary_integer
;

type t_num_tbl is
table of number
index by binary_integer
;

   -- selects entities linked to business rules with remark that indicate usage of
   -- relationship
   cursor c_funent
   (      b_fun_id in number
   )      is
   select fent.entity_reference
   from   ci_function_entity_usages     fent
   where  fent.function_reference     = b_fun_id
   and    fent.retrieve_flag           = 'Y'
   ;

   -- selects foreign key column references that references the given relationship
   cursor c_fco_col
   (      b_relend_id in number
   ,      b_tbl_id    in number
   ) is
   select oky.column_reference
   from   ci_key_components          oky
   ,      ci_foreign_key_constraints fco
   where  fco.relationship_end_reference = b_relend_id
   and    oky.constraint_reference       = fco.id
   and    fco.table_reference            = b_tbl_id
   ;

   -- 6.5.3.4AMIS
   --  selects the column(s) that are in the primary key of the table that is created
   -- for the Entity in whose BR_ENT-uasage the <PK> tag was contained.
   cursor   c_pk_use
   (        b_fun_id       in number
   ,        b_ent_id        in number
   ,        b_tbl_id        in number
   )        is
   select   oky.column_reference col_id
   ,        fent.remark
   ,        col.name col_name
   from     ci_function_entity_usages  fent
   ,        ci_key_components          oky
   ,        ci_primary_key_constraints pco
   ,        ci_columns                 col
   where    fent.function_reference         =  b_fun_id
   and      fent.entity_reference           =  b_ent_id
   and      instr(fent.remark, PRIMARY_KEY_TAG) != 0
   and      oky.constraint_reference       = pco.id
   and      pco.table_reference            = b_tbl_id
   and      oky.column_reference           = col.id
   and      col.source_attribute_reference is null -- only retrieve Surrogate Key Columns that are not available as attribute in the Entity Definition
   ;

   
g_trg_tbl          t_trg_tbl_tbl;
g_trg_col          t_trg_col_tbl;
g_transition_rule  boolean;
g_delete_rule      boolean;

e_user_error       exception;

function revision
return varchar2
is
begin
   return blvers.get_revision_label(REVISION_LABEL);
end revision;

procedure install
is
begin
   hsu_install.add_utility
   ( PACKAGE_NAME
   , UTILITY_NAME
   , 'PB'
   , 'This Productivity Booster is part of a suite of utilities in support of CDM '
   ||'RuleFrame.  See the CDM Standards and Guidelines Library'
   ||' and the Headstart User Guide for more information.'
   ||chr(10)
   ||chr(10)
   ||'This utility creates Business Rule Design Definitions'
   ||' for the atomic functions among the specified Business Rule Functions.'
   ||' If a Design Definition already exists for a certain Rule,'
   ||' the Rule is skipped to prevent overwriting.'
   ||chr(10)
   ||chr(10)
   ||' Select the Business Rule Functions you want to transform, based on '
   ||'<B>one</B> of the following parameters:'
   ||chr(10)||' Business Rule Function'
   ||chr(10)||' Filter for BR Functions'
   ||chr(10)||' BR Function Tree Root'
   ||chr(10)||' Rules triggered by Entity'
   ||chr(10)
   ||'The label of the Business Rule Function is used to determine the Rule Type.'
   ||' Apply CDM standard OMS-60005 to make sure that the'
   ||' created Design Definitions get the correct Rule Type'
   ||' (for example, the function label must start with ''BR'').'
   ||chr(10)
   ||chr(10)
   ||'The utility can only transform Business Rule Functions that have'
   ||' triggering events that satisfy the following conditions.'
   ||chr(10)||'<ul><li>Event type is Change</li>'
   ||chr(10)||'<li>On Condition is CREATE, UPDATE or DELETE</li>'
   ||chr(10)||'<li>Entity property is specified</li></ul>'
   ||chr(10)||'Possibly also the Attribute property is specified'
   ||' or the relationship is indicated in the Description using the convention'
   ||' [relationship name]->[other end entity name].'
   ||chr(10)||chr(10)
   ||'There is one exception: if the Rule Type is Static'
   ||' and the Business Rule Function does not have any'
   ||' triggering Events, it is assumed that'
   ||' each Entity/Attribute usage also represents a Create and Update triggering Event.'
   ||chr(10)
   ||chr(10)
   ||'An additional condition is that each triggering Entity is implemented by '
   ||' preferably one Table (the first is taken),'
   ||' which must be accessible to the user running the utility.'
   ||' If the (first) table implementing the entity is in a folder which is not'
   ||' accessible to the user, you get the error'
   ||' CDA-06000: Access control violation: not found or no rights on this application system.'
   ||chr(10)
   ||chr(10)
   ||' The Entity/Attribute Retrieve usages are transformed to Arguments'
   ||' of the Design Definitions.'
   ||' Specify Relationship usages by adding one or more of the following'
   ||' tags in the comment of the relevant Entity usage:'
   ||htf.escape_sc(' <RL>[relationship name]->[other end entity name]</RL>.')
   ||'You can also specify to the transformer that'
   ||' Attribute/Relationship usages must be created'
   ||' based on the triggering Attributes/Relationships.'
   ||chr(10)
   ||chr(10)
   ||'The Transformer will create a default message for each Business Rule,'
   ||' containing the Function Short Definition as its message text.'
   );

   hsu_install.add_parameter
   ( PACKAGE_NAME                         -- package name
   , 10                                   -- sequence
   , 'Business Rule Function'             -- prompt
   , null                                 -- default value actual
   , '<null>'                             -- default value displayed
   , 'N'                                  -- mandatory
   , 'Y'                                  -- allow multi-select
   , 'N'                                  -- include shared elements
   , 'FUN'                                -- element type short name
   , 'TO_CHAR(ID)'                        -- sql expression actual
   , 'FUNCTION_LABEL'                     -- sql expression displayed
   , 'hsu_name.is_business_rule(fun.function_label) = ''Y'''  -- 6.5.0.0  check br label via function
     ||' and hsu_brtr.is_atomic(fun.irid) = ''Y'''            -- 6.5.0.0  atomic via function
                                          -- where clause
   , null                                 -- synchronize with
   , null                                 -- foreign key column
   ,                                      -- help text
     'Choose the Business Rule Functions that should be transformed.'
     ||' Only atomic functions whose label starts with BR are shown.'
   );

   hsu_install.add_parameter
   ( PACKAGE_NAME                         -- package name
   , 15                                   -- sequence
   , 'Filter for BR Functions'            -- prompt
   , null                                 -- default value actual
   , ''                                   -- default value displayed
   , 'N'                                  -- mandatory
   , 'N'                                  -- allow multi-select
   , 'N'                                  -- include shared elements
   , ''                                   -- element type short name
   , ''                                   -- sql expression actual
   , ''                                   -- sql expression displayed
   , null                                 -- where clause
   , null                                 -- synchronize with
   , null                                 -- foreign key column
   ,                                      -- help text
     'Define a name filter for Business Rule Functions (such as BR_%_CEW, or BR_PRJ%).'
     ||' If you want to select all Business Rule Functions, use %.'
   );

   hsu_install.add_parameter
   ( PACKAGE_NAME                         -- package name
   , 20                                   -- sequence
   , 'BR Function Tree Root'              -- prompt
   , null                                 -- default value actual
   , '<null>'                             -- default value displayed
   , 'N'                                  -- mandatory
   , 'N'                                  -- allow multi-select
   , 'N'                                  -- include shared elements
   , 'FUN'                                -- element type short name
   , 'TO_CHAR(ID)'                        -- sql expression actual
   , 'FUNCTION_LABEL'                     -- sql expression displayed
   , 'hsu_name.is_business_rule(fun.function_label) = ''Y'''
                                          -- 6.5.0.0   check br label via function
                                          -- where clause
   , null                                 -- synchronize with
   , null                                 -- foreign key column
   ,                                      -- help text
     'Choose the Root Business Rule Function'
     ||' for the tree of Business Rule Functions that should be transformed.'
     ||' Only functions whose label starts with BR are shown.'
   );

   hsu_install.add_parameter
   ( PACKAGE_NAME                         -- package name
   , 25                                   -- sequence
   , 'Rules triggered by Entity'          -- prompt
   , null                                 -- default value actual
   , '<null>'                             -- default value displayed
   , 'N'                                  -- mandatory
   , 'Y'                                  -- allow multi-select
   , 'Y'                                  -- include shared elements
   , 'ENT'                                -- element type short name
   , 'TO_CHAR(ID)'                        -- sql expression actual
   , 'NAME'                               -- sql expression displayed
   , null                                 -- where clause
   , null                                 -- synchronize with
   , null                                 -- foreign key column
   ,                                      -- help text
     'Choose the Entities that trigger the Business Rule Functions'
     ||' that should be transformed.'
   );


   hsu_install.add_parameter
   ( PACKAGE_NAME                         -- package name
   , 27                                   -- sequence
   , 'Create Attr/Relationship Usages'           -- prompt
   , 'N'                                  -- default value actual
   , 'No'                                 -- default value displayed
   , 'Y'                                  -- mandatory
   , 'N'                                  -- allow multi-select
   , 'N'                                  -- include shared elements
   , ''                                   -- element type short name
   , ''                                   -- sql expression actual
   , ''                                   -- sql expression displayed
   , null                                 -- where clause
   , null                                 -- synchronize with
   , null                                 -- foreign key column
   ,                                      -- help text
    'Choose whether you want to create Attribute usages '
    ||' and Relationship usage tags for the triggering Attributes'
    ||' and Relationships in the Events.'
    ||' If you set this parameter to Yes, a retrieve usage will be created'
    ||' for any triggering attribute or relationship, '
    ||' even if the rule already has other retrieve usages.'
   );

   hsu_install.add_allowable_value
   ( PACKAGE_NAME                 -- package name
   , 27                           -- parameter
   , 1                            -- sequence
   , 'Y'                          -- actual value
   , 'Yes'                        -- displayed value
   );

   hsu_install.add_allowable_value
   ( PACKAGE_NAME                 -- package name
   , 27                           -- parameter
   , 2                            -- sequence
   , 'N'                          -- actual value
   , 'No'                         -- displayed value
   );


   hsu_install.add_parameter -- 6.5.3.3 new
   ( PACKAGE_NAME                         -- package name
   , 28                                   -- sequence
   , 'Create constr. for ATT/TPL rules'   -- prompt
   , 'Y'                                  -- default value actual
   , 'Yes'                                -- default value displayed
   , 'Y'                                  -- mandatory
   , 'N'                                  -- allow multi-select
   , 'N'                                  -- include shared elements
   , ''                                   -- element type short name
   , ''                                   -- sql expression actual
   , ''                                   -- sql expression displayed
   , null                                 -- where clause
   , null                                 -- synchronize with
   , null                                 -- foreign key column
   ,                                      -- help text
    'Choose whether you want to implement Attribute and Tuple rules '
    ||' in database check constraints instead of in the CAPI.'
    ||' Check constraints have a better performance, but if a'
    ||' violation occurs the end user gets one message at a time,'
    ||' instead of getting all messages in a list as with the CAPI.'
    ||chr(10)
    ||chr(10)
    ||'Some ATT/TPL rules cannot be implemented in Check Constraints'
    ||' because the validation condition cannot be expressed in SQL'
    ||' (without constructs that are forbidden in check constraints).'
    ||' If that is the case, run the Transformer separately for those'
    ||' rules and set this parameter to No.'
   );

   hsu_install.add_allowable_value
   ( PACKAGE_NAME                 -- package name
   , 28                           -- parameter
   , 1                            -- sequence
   , 'Y'                          -- actual value
   , 'Yes'                        -- displayed value
   );

   hsu_install.add_allowable_value
   ( PACKAGE_NAME                 -- package name
   , 28                           -- parameter
   , 2                            -- sequence
   , 'N'                          -- actual value
   , 'No'                         -- displayed value
   );

-- 6.5.1.1   altered the help text to mention hs4apps
   hsu_install.add_parameter
   ( PACKAGE_NAME                 -- package name
   , 30                           -- sequence
   , 'Message Prefix'             -- prompt
   , '<default>'                  -- default value actual
   , '<default>'                  -- default value displayed -- 6.5.0.0  changed to <default>
   , 'Y'                          -- mandatory
   , 'N'                          -- allow multi-select
   , 'N'                          -- include shared elements
   , ''                           -- element type short name
   , ''                           -- sql expression actual
   , ''                           -- sql expression displayed
   , null                         -- where clause
   , null                         -- synchronize with
   , null                         -- foreign key column
   ,  'Specify the prefix for the code of the message that should be displayed upon'
    ||' violation of this business rule. This prefix should be the application code.'
    ||' If you are generating messages for HS4APPS, this prefix should be the'
    ||' application code as defined in AOL.'||chr(10)
    ||chr(10)
    ||'If the value is '''||htf.escape_sc('<default>')
    ||''', the utility will take the three letter abbreviation of the owning container '
    ||'where the pl/sql module for the design definition is created.'
                                  -- help text
   );


   hsu_install.add_msg_language_parameter
   ( PACKAGE_NAME                 -- package name
   , 35                           -- sequence
   , 'Message Language'           -- prompt
   , -- help text
     'Specify the language that the default messages must be created in.'
   );
   -- 6.5.3.4AMIS
      hsu_install.add_parameter
   ( PACKAGE_NAME                         -- package name
   , 40                                   -- sequence
   , 'Overwrite existing BRDDs?'           -- prompt
   , 'N'                                  -- default value actual
   , 'No'                                 -- default value displayed
   , 'Y'                                  -- mandatory
   , 'N'                                  -- allow multi-select
   , 'N'                                  -- include shared elements
   , ''                                   -- element type short name
   , ''                                   -- sql expression actual
   , ''                                   -- sql expression displayed
   , null                                 -- where clause
   , null                                 -- synchronize with
   , null                                 -- foreign key column
   ,                                      -- help text
    'Choose whether you want this utility to overwrite existing Business Rule Design Definitions.'
   );
   hsu_install.add_allowable_value
   ( PACKAGE_NAME                 -- package name
   , 40                           -- parameter
   , 1                            -- sequence
   , 'Y'                          -- actual value
   , 'Yes'                        -- displayed value
   );

   hsu_install.add_allowable_value
   ( PACKAGE_NAME                 -- package name
   , 40                           -- parameter
   , 2                            -- sequence
   , 'N'                          -- actual value
   , 'No'                         -- displayed value
   );
   --------END of 6.5.3.4AMIS1.1

end install;

procedure deinstall
is
begin
   hsu_install.deinstall(PACKAGE_NAME);
end deinstall;

-- 6.5.0.0  new function, atomic flag is removed in 6i
function is_atomic
( p_fun_irid in ci_functions.irid%type
)
return varchar2
--
-- Purpose  returns Y if function has no child functions, N otherwise
--
-- Usage    from hsu_brtr.install (parameter where clause)
--
-- Remarks  -
--
is
   -- cursor to select child records for a given function
   cursor c_fun
   ( b_fun_irid in ci_functions.irid%type
   )
   is
   select fun.irid
   from   ci_functions fun
   where  fun.parent_reference = b_fun_irid
   ;
   l_fun_irid ci_functions.irid%type;
begin
   open c_fun(p_fun_irid);
   fetch c_fun into l_fun_irid;
   if c_fun%found
   then
      close c_fun;
      return 'N';
   else
      close c_fun;
      return 'Y';
   end if;
end is_atomic;

-- 6.5.1.0  removed function fun_label, now hsu_name.get_name_and_path is used
-- 6.5.1.0  removed function ent_name, now hsu_name.get_name_and_path is used
-- 6.5.1.0  removed function ent_short_name: never used
-- 6.5.1.0  removed function ent_plural: never used
-- 6.5.1.0  removed function tbl_name, now hsu_name.get_name_and_path is used

function att_name
( p_att_id in number
) return varchar2
is
--
-- Purpose  This function determines the name of the attribute based on the attribute id.
--
-- Usage    p_att_id : is the id of the attribute
--
-- Remarks  -
--
   l_string varchar2(50);
begin
   select att.name
   into   l_string
   from   ci_attributes att
   where  att.id = p_att_id
   ;
   return l_string;
exception
   when no_data_found
   then
      return null;
end; -- att_name

function relend_name
( p_relend_id in number
) return varchar2
is
--
-- Purpose  This function determines the relationship name based on the relationship id.
--
-- Usage    p_relend_id : is the id of the relationship
--
-- Remarks  -
--
   l_string varchar2(50);
begin
   select relend.name||'->'||to_ent.name
   into   l_string
   from   ci_relationship_ends relend
   ,      ci_entities to_ent
   where  relend.id = p_relend_id
   and    to_ent.id = relend.to_entity_reference
   ;
   return l_string;
exception
   when no_data_found
   then
      return null;
end; -- relend_name

function col_name
( p_col_id in number
) return varchar2
is
--
-- Purpose  This function determines the name of the column based on the column id.
--
-- Usage    p_col_id : is the id of the column
--
-- Remarks  -
--
   l_string varchar2(50);
begin
   select col.name
   into   l_string
   from   ci_columns col
   where  col.id = p_col_id
   ;
   return l_string;
exception
   when no_data_found
   then
      return null;
end; -- col_name


procedure add_to_num_tbl
( p_num     in     number
, p_num_tbl in out t_num_tbl
)
is
--
-- Purpose  This procedure adds one element p_num to the p_num_tbl
--
-- Usage    p_num     : is the actual number that should be added to the p_num_tbl
--          p_num_tbl : is the number table to where the value should be added
--
-- Remarks  -
--
--
begin
   p_num_tbl( p_num_tbl.count + 1):= p_num;
end; -- add_to_numi_tbl

function exists_in_num_tbl
( p_num in number
, p_num_tbl in t_num_tbl
) return boolean
is
--
-- Purpose  this function indicates whether element p_num exists in the p_num_tbl
--
-- Usage    p_num     : is the number that should is checked to exist in p_num_tbl
--          p_num_tbl : is the number table to where the value may exist
--
-- Remarks  -
--
--
begin
   for i in 1..p_num_tbl.count loop
      if p_num_tbl(i) = p_num
      then
         return true;
      end if;
   end loop;
   return false;
end; -- add_to_numi_tbl


function find_col
( p_tbl_id in number
, p_att_id in number
) return number
is
--
-- Purpose  This function determines the column mapped from a given attribute and table
--
-- Usage    p_tbl_id : is the table id of the column that should be determined
--          p_att_id : is the attribute id to which the column is mapped
--
-- Remarks  -
--
--
   cursor c_col
   ( b_tbl_id in ci_table_definitions.id%type
   , b_att_id in ci_attributes.id%type
   )
   is
   select col.id
   from   ci_columns col
   where  col.table_reference            = b_tbl_id
   and    col.source_attribute_reference = b_att_id
   ;
   r_col c_col%rowtype;
   --
   cursor c_att
   ( b_att_id in ci_attributes.id%type
   )
   is
   select att.entity_reference
   ,      att.name
   from   ci_attributes att
   where  att.id = b_att_id
   ;
   r_att c_att%rowtype;
begin
   -- 6.5.0.0  function changed: raise error when column not found
   open c_col(p_tbl_id, p_att_id);
   fetch c_col
   into r_col;
   if c_col%found
   then
      close c_col;
      return r_col.id;
   else
      close c_col;
      -- column attribute reference not found, give error and abort utility
      open c_att(p_att_id);
      fetch c_att
      into  r_att;
      close c_att;
      bllog.write
      ( 'Column reference for '||hsu_name.get_name_and_path(r_att.entity_reference)
        ||'.'||r_att.name||' not found'
      , bllog.errors
      );
      raise e_user_error;
   end if;
end; -- find_col


function find_relend_id
( p_from_entity     in  number
, p_description     in  varchar2
-- 6.5.1.0  new parameter p_evt_id
, p_evt_id          in  number
, p_evt_name        in  varchar2
, p_to_entity       out number
, p_other_relend_id out number
, p_superent_id     out number
) return number
is
--
-- Purpose  This function determines id of the relationship dependent on
--          the event description or the event name.
--          It also returns (through out parameters) the id of the entity to which the
--          relationship is defined, the id of the relationship defined the other way
--          around, and if the relationship is defined for a subentity, but is actually
--          directed to the superentity, then the id of the superentity is also given.
--
-- Usage    p_from_entity : is the id of the entity from which the relationship is defined.
--          p_desciption  : is the event description
--          p_evt_id       : is the id of the event
--          p_evt_name     : is the name of the event
--          p_to_entity     : is the name of the entity to which the relationship is defined
--          p_other_relend_id : the id of the relationship defined the other way around.
--          p_superent_id : is the id of the superentity if appropriate.
--
-- Remarks  -
--
   cursor c_relend_desc
   (      b_from_ent_id  in number
   ,      b_description  in varchar2
   )      is
   select relend.id                         relend_id
   ,      to_ent.id                         to_ent_id
   ,      relend.other_relationship_end_ref other_relend_id
   from   ci_relationship_ends relend
   ,      ci_entities          to_ent
   where  relend.from_entity_reference = b_from_ent_id
   and    relend.to_entity_reference   = to_ent.id
   and    instr( upper( b_description)
               , upper( relend.name
                        ||'->'
                        ||to_ent.name
                      )
               ) > 0
   ;

   l_relend_id       number(38);
   l_other_relend_id number(38);
   l_to_ent_id       number(38);
   l_superent_id     number(38);

begin
   if p_description is not null
   then
      <<relend_desc>>
      for r_relend_desc in c_relend_desc
      ( p_from_entity
      , p_description
      ) loop
         l_relend_id       := r_relend_desc.relend_id;
         l_to_ent_id       := r_relend_desc.to_ent_id;
         l_other_relend_id := r_relend_desc.other_relend_id;
      end loop relend_desc;

      if l_relend_id is null -- check subtype
      then
         -- check whether the entity is a sub-entity, and if that is the case, then
         -- check whether the relationship belongs to the super entity
         <<supertype>>
         for c_ent in blent.c_ent(p_from_entity) loop
            if c_ent.ent_supertype_reference is not null
            then
               bllog.write
               ( 'Entity has supertype reference.  Checking whether '
                 ||'relationship belongs to supertype.'
               , bllog.debug
               );

               l_relend_id:= find_relend_id
                             ( p_from_entity     => c_ent.ent_supertype_reference
                             , p_description     => p_description
                             , p_evt_id          => p_evt_id
                             , p_evt_name        => p_evt_name
                             , p_to_entity       => l_to_ent_id
                             , p_other_relend_id => l_other_relend_id
                             , p_superent_id     => l_superent_id
                             );
               if l_relend_id is not null
               then
                  l_superent_id := c_ent.ent_supertype_reference;
               end if;
            end if; -- supertype_reference is not null
         end loop supertype;
      end if; --l_relend_id is null (check subtype)

      if l_relend_id is null -- no subtype
      then
         bllog.write
         ( 'Relationship not identified in find_relend_id.'
           ||' for event_name = '
           ||p_evt_name
           ||' and description = '
           ||p_description
         , bllog.debug
         );
      end if; -- l_relend_id is null (no subtype)

   elsif instr( p_evt_name
              , '->'
              ) > 0
   then
      select relend.id
      ,      to_ent.id
      ,      relend.other_relationship_end_ref
      into   l_relend_id
      ,      l_to_ent_id
      ,      l_other_relend_id
      from   ci_relationship_ends relend
      ,      ci_entities          to_ent
      where  relend.from_entity_reference = p_from_entity
      and    relend.to_entity_reference   = to_ent.id
      -- check if in the part of the event name starting after the . at least part of the
      -- proper relationship indication (name->to_entity_name) can be found
      and    instr( upper( p_evt_name)
                  , substr
                    ( upper(relend.name||'->'||to_ent.name)
                    , 1
                    , length(p_evt_name) - instr(p_evt_name, '.')
                    )
                  ) > 0
      ;
   end if;
   p_to_entity       := l_to_ent_id;
   p_other_relend_id := l_other_relend_id;
   p_superent_id     := l_superent_id;
   if l_relend_id is null
   then
      raise e_user_error;
   end if;
   return l_relend_id;
exception
   when others
   then
      bllog.write
      ( 'Relationship can not be identified for Event '
        -- 6.5.1.0  get name and path
        ||hsu_name.get_name_and_path(p_evt_id)
        ||' with Description = '
        ||p_description
      , bllog.errors
      );
      raise e_user_error;
      return null;
end find_relend_id;

-- 6.5.2.2
function relation_as_fk
( p_relend_id       in number
, p_from_ent_id     in number
, p_to_ent_id       in number
, p_other_relend_id in number
) return boolean
is
--
-- Purpose  This function determines whether the given relationship has been implemented
--          as a Foreign Key
--
-- Usage    p_relend_id       : is the id of the relationship seen from this end
--          p_from_ent_id     : is the id of the entity at this end
--          p_to_ent_id       : is the id of the entity at the other end of the relationship
--          p_other_relend_id : is the id of the relationship seen from the other end
--
--          When a FK is found, then the function returns TRUE.
--
--
-- Remarks  -
--
--
   l_test varchar2(1);
begin
   -- non-owned self-referencing relationships are not correctly recognized!
   bllog.write
   ( ' relation_as_fk checked for relationship end: '
     ||to_char(p_relend_id)
     ||' '
     ||relend_name(p_relend_id)
     ||' in the context of (from_)entity '
     ||to_char(p_from_ent_id)
     ||' '
     ||hsu_name.get_name_and_path(p_from_ent_id)
   , bllog.debug
   );

   begin -- find fco
      select 'x'
      into   l_test
      from   dual
      where  exists
             ( select 'x'
               from   ci_tables_entities          tabent
               ,      ci_foreign_key_constraints  fco
               where  fco.relationship_end_reference = p_relend_id
               and    fco.table_reference = tabent.table_reference
               and    tabent.entity_reference = p_from_ent_id
             )
      ;
      bllog.write
      ( 'found foreign key - probably owner'
      , bllog.debug
      );
      return true;
   exception
      when no_data_found
      then
         -- find a Foreign Key implementing the Other Relend referencing (foreign table)
         -- a Table mapping the Entity
         select 'x'
         into   l_test
         from   dual
         where  exists
                ( select 'x'
                  from   ci_tables_entities          tabent
                  ,      ci_foreign_key_constraints  fco
                  where  fco.relationship_end_reference = p_other_relend_id
                  and    fco.foreign_table_reference    = tabent.table_reference
                  and    tabent.entity_reference        = p_from_ent_id
                )
         ;
         -- when found then a non-owned relend
         bllog.write
         ( 'found foreign key - not owner'
         , bllog.debug
         );
         return true;
   end; -- find_fco
exception
   when no_data_found
   then
      -- no fk was found
      -- the relend has not been implemented at all?
      bllog.write
      ( 'function relation_as_fk: NO_DATA_FOUND for relationship end: '
        ||to_char(p_relend_id)
        ||' '
        ||relend_name(p_relend_id)
        ||' in the context of (from_)entity '
        ||to_char(p_from_ent_id)
        ||' '
        ||hsu_name.get_name_and_path(p_from_ent_id)
      , bllog.debug
      );
      bllog.write
      ( '<B>No Foreign Key found that implements the relationship end '
        ||relend_name(p_relend_id)
        ||' of entity '||hsu_name.get_name_and_path(p_from_ent_id)
        ||'; Create the FK or fill in the Source Relationship property of the FK.'
        ||chr(10)
        ||'The created BRDD has to be adjusted manually if the relationship'
        ||' is implemented as Domain Key.</B>'
      , bllog.information
      );
      return false;
end relation_as_fk;
-- 6.5.2.2

function relend_owned
( p_relend_id       in number
, p_from_ent_id     in number
, p_to_ent_id       in number
, p_other_relend_id in number
) return boolean
is
--
-- Purpose  This function determines whether the given relationship has been implemented
--          as a Foreign Key from a table mapped to the other end entity, referencing a
--          table mapped to the entity at this end
--
-- Usage    p_relend_id       : is the id of the relationship seen from this end
--          p_from_ent_id     : is the id of the entity at this end
--          p_to_ent_id           : is the id of the entity at the other end of the relationship
--          p_other_relend_id : is the id of the relationship seen from the other end
--
--          When the relend is owned by the given from entity, then the function returns TRUE.
--          When it is owned by the entity at the other end, then the function returns FALSE
--
--
-- Remarks  -
--
--
   l_test varchar2(1);
begin
   -- non-owned self-referencing relationships are not correctly recognized!
   bllog.write
   ( ' relend_owned checked for relationship end: '
     ||to_char(p_relend_id)
     ||' '
     ||relend_name(p_relend_id)
     ||' in the context of (from_)entity '
     ||to_char(p_from_ent_id)
     ||' '
     ||hsu_name.get_name_and_path(p_from_ent_id)
   , bllog.debug
   );

   begin -- find fco
      -- find a Foreign implementing the Relend owned by a Table mapping the Entity
      select 'x'
      into   l_test
      from   dual
      where  exists
             ( select 'x'
               from   ci_tables_entities          tabent
               ,      ci_foreign_key_constraints  fco
               where  fco.relationship_end_reference = p_relend_id
               and    fco.table_reference = tabent.table_reference
               and    tabent.entity_reference = p_from_ent_id
             )
      ;
      bllog.write
      ( 'found foreign key - probably owner'
      , bllog.debug
      );
      -- when found then probably an owned relend (unless self-referencing)
      return true;
   exception
      when no_data_found
      then
         -- find a Foreign implementing the Other Relend referencing (foreign table)
         -- a Table mapping the Entity
         select 'x'
         into   l_test
         from   dual
         where  exists
                ( select 'x'
                  from   ci_tables_entities          tabent
                  ,      ci_foreign_key_constraints  fco
                  where  fco.relationship_end_reference = p_other_relend_id
                  and    fco.foreign_table_reference    = tabent.table_reference
                  and    tabent.entity_reference        = p_from_ent_id
                )
         ;
         -- when found then a non-owned relend
         bllog.write
         ( 'found foreign key - not owner'
         , bllog.debug
         );
         return false;
   end; -- find_fco
exception
   when no_data_found
   then
      -- no owned relend was found, but no non-owned either
      -- perhaps the relend has not been implemented at all?
      bllog.write
      ( 'function relend_owned: NO_DATA_FOUND for relationship end: '
        ||to_char(p_relend_id)
        ||' '
        ||relend_name(p_relend_id)
        ||' in the context of (from_)entity '
        ||to_char(p_from_ent_id)
        ||' '
        ||hsu_name.get_name_and_path(p_from_ent_id)
      , bllog.debug
      );
      -- 6.5.1.2  fix for bug 1603181: instead of just return false
      -- (which would indicate that the other end is the owner), raise an error
      bllog.write
      ( 'Cannot process triggering Event for Relationship End '
        ||relend_name(p_relend_id)
        ||' of Entity '||hsu_name.get_name_and_path(p_from_ent_id)
        ||'; no Foreign Key found that implements this relationship.'
        ||' Create the FK or fill in the Source Relationship property of the FK.'
      , bllog.errors
      );
      raise e_user_error;
end relend_owned;


procedure check_relend_event
( p_evt_id           in     ci_events.id%type
, p_evt_name         in     ci_events.name%type
, p_ent_id           in out ci_entities.id%type
, p_relend_id           out ci_relationship_ends.id%type
, p_to_ent_id           out ci_entities.id%type
, p_other_relend_id     out ci_relationship_ends.id%type
, p_superent_id         out ci_entities.id%type
, p_non_owned_relend    out boolean
)
--  Purpose  Fill the out parameters with the right relationship info
--           if the event is a relationship event
--
is
   l_text_lines       bltext.cdi_text_lines;
   l_text_count       number(5);
   l_relend_descr     varchar2(500);
   -- local variables to fill output parameters
   l_ent_id           ci_entities.id%type;
   l_relend_id        ci_relationship_ends.id%type;
   l_to_ent_id        ci_entities.id%type;
   l_other_relend_id  ci_relationship_ends.id%type;
   l_superent_id      ci_entities.id%type;
   l_non_owned_relend boolean := false;
begin
   l_ent_id := p_ent_id;
   -- Check whether the event refers to updating a relationship
   -- get relationship from the Event Description
   bltext.get_lines_from_repository
   ( p_evt_id
   , 'CDIDSC'
   , l_text_lines
   , l_text_count
   );
   l_relend_descr := bltext.get_html_tagged_text  -- 6.5.0.2  changed
                     ( l_text_lines
                     , l_text_count
                     , 'RL'
                     );
   -- check the relationship description
   if l_relend_descr is not null
   or instr(p_evt_name,'->') > 0
   then
      -- the triggering event refers to a relationship
      -- determine the relend_id
      l_relend_id:= find_relend_id
                    ( p_from_entity     => l_ent_id
                    , p_description     => l_relend_descr
                    , p_evt_id          => p_evt_id
                    , p_evt_name        => p_evt_name
                    , p_to_entity       => l_to_ent_id
                    , p_other_relend_id => l_other_relend_id
                    , p_superent_id     => l_superent_id
                      );
      -- 6.5.2.2 check if this relationship corresponds with a foreign key
      -- relationships that are not implemented are listed in the log file
      -- no error is raised because relationship might be implemented as domain key
      if not relation_as_fk( p_relend_id       => l_relend_id
                           , p_from_ent_id     => l_ent_id
                           , p_to_ent_id       => l_to_ent_id
                           , p_other_relend_id => l_other_relend_id
                           )
      then
        l_to_ent_id := null;
        l_non_owned_relend := false;
      else
      -- 6.5.2.2
        -- check if this entity is the 'owning' entity
        -- that is, will the FK be created in this table
        l_non_owned_relend := not relend_owned
                              ( p_relend_id      => l_relend_id
                              , p_from_ent_id    => l_ent_id
                              , p_to_ent_id        => l_to_ent_id
                              , p_other_relend_id => l_other_relend_id
                              );
        if l_non_owned_relend
        then
           -- the relationship has been implemented as a Foreign Key from a table
           -- mapped to the other entity, referencing a table mapped to this entity
           l_ent_id:= l_to_ent_id; -- find/work from other entity
        end if; -- non-owned
     end if; -- 6.5.2.2
   end if; -- check relationship description
   -- set the output parameters
   p_ent_id := l_ent_id;
   p_relend_id := l_relend_id;
   p_to_ent_id := l_to_ent_id;
   p_other_relend_id := l_other_relend_id;
   p_superent_id := l_superent_id;
   p_non_owned_relend := l_non_owned_relend;
end check_relend_event;


function find_discriminator_col_id
( p_tbl_id in number
) return number
is
--
-- Purpose  This function determines id of the discriminator column for a given table
--
-- Usage    p_tbl_id: id of the table where the discriminator column should be determined.
--
-- Remarks  -
--
   l_disc_col number(38);
begin
   select tabent.column_reference
   into   l_disc_col
   from   ci_tables_entities tabent
   where  tabent.table_reference = p_tbl_id
   and    tabent.column_reference is not null
   and    rownum = 1
   ;
   return l_disc_col;
exception
   when no_data_found
   then
      return null;
end find_discriminator_col_id;


procedure add_trg_col
( p_tbl_id     in number
, p_col_id     in number
) is
--
-- Purpose  This procedure adds a given column to the list of columns that should
--          be included as triggering columns for the business rule.
--
-- Usage    p_tbl_id : id of the table the column belongs to
--          p_col_id : id of the column that should be added to the list
--
-- Remarks  -
--
   l_trg_col_index number(5):= g_trg_col.count + 1;
begin
   <<count>>
   for i in 1..g_trg_col.count loop
      if g_trg_col(i).col_id = p_col_id
      then
         l_trg_col_index:= i;
         exit;
      end if;
   end loop count;
   g_trg_col(l_trg_col_index).tbl_id := p_tbl_id;
   g_trg_col(l_trg_col_index).col_id := p_col_id;
end add_trg_col;


procedure add_tbl
( p_tbl_id    in number
, p_event     in varchar2
, p_when      in varchar2 default null
) is
--
--
-- Purpose  This procedure adds a table to the list of tables that trigger the business
--            rule.  If the tables does not yet exist in the list (g_trg_tbl), then
--            it is added.  If it exists, it is updated with the provided information in
--            the following way:
--          - the event-boolean (insert,update,delete) is set to true whenever appropriate;
--            the others are left as is
--          - the when condition is extended with 'AND '||p_when when a when condition is
--            provided
--
-- Usage    p_tbl_id : id of the table to include or update with information
--          p_event  : is CREATE, INSERT, UPDATE, DELETE; case insensitive,
--                     validation on first character only
--          p_when   : is a when condition indicating when the trigger should apply
--
-- Remarks  -
--
   l_trg_tbl_index number(5):= g_trg_tbl.count + 1;
begin
   <<count>>
   for i in 1..g_trg_tbl.count loop
      if g_trg_tbl(i).tbl_id= p_tbl_id
      then
         l_trg_tbl_index:= i;
         exit;
      end if;
   end loop count;

   g_trg_tbl( l_trg_tbl_index).tbl_id := p_tbl_id;
   if upper(substr(p_event, 1,1)) in ('I', 'C')
   or p_event = ON_CONDITION_CREATE
   then
      g_trg_tbl( l_trg_tbl_index).ins:= true;
   end if;

   if upper(substr(p_event, 1,1)) ='U'
   or p_event = ON_CONDITION_UPDATE
   then
      g_trg_tbl( l_trg_tbl_index).upd:= true;
   end if;

   if upper(substr(p_event, 1,1)) ='D'
   or p_event = ON_CONDITION_DELETE
   then
      g_trg_tbl( l_trg_tbl_index).del:= true;
   end if;

   if p_when is not null
   then
      if g_trg_tbl( l_trg_tbl_index).when_condition is null
      then
         g_trg_tbl( l_trg_tbl_index).when_condition:= p_when;
      else
         -- check that the when condition to be added
         -- is not part of the existing when condition yet
         if instr(g_trg_tbl(l_trg_tbl_index).when_condition, p_when) = 0
         then
            -- add the new when condition
            g_trg_tbl( l_trg_tbl_index).when_condition :=
                    g_trg_tbl( l_trg_tbl_index).when_condition
                    -- 6.0.3.1 multiple subtypes: or instead of and
                    ||' or '
                    ||p_when
                    ;
         end if;
      end if; -- g_trg_tbl( l_trg_tbl_index).when_condition is null
   end if; -- p_when is not null
end add_tbl;


procedure create_att_usage
( p_fun_id   in number
, p_ent_id   in number
, p_att_id   in number
) is
--
-- Purpose  This procedure creates attribute usages for the given function and entity
--            when it does not already exist
--
-- Usage    p_fun_id: id of the function to which the usage should be added
--            p_ent_id: id of the entity to which the attribute belongs
--            p_att_id: id of the attribute that should be added as usage to the function
--
-- Remarks  -
--

   -- checks whether there is an existing entity usage
   cursor c_funent
   (      b_fun_id in number
   ,      b_ent_id in number
   )      is
   select 'x'
   from   ci_function_entity_usages  fent
   where  fent.function_reference  = b_fun_id
   and    fent.entity_reference    = b_ent_id
   ;

   -- checks whether there is an existing attribute usage
   cursor c_funatt
   (      b_fun_id in number
   ,      b_ent_id in number
   ,      b_att_id in number
   )      is
   select 'x'
   from   ci_function_attribute_usages  fatt
   where  fatt.function_reference  = b_fun_id
   and    fatt.entity_reference    = b_ent_id
   and    fatt.attribute_reference = b_att_id
   ;

   l_funatt_exists boolean := false;
   l_funent_exists boolean := false;
   l_funent_id     ci_function_entity_usages.id%type;
   l_funatt_id     ci_function_attribute_usages.id%type;

begin

   bllog.write
   ('Create Function Entity Usage if it does not already exist.'
   , bllog.debug
   );

   <<funent>>
   for r_funent in c_funent
                   ( b_fun_id => p_fun_id
                   , b_ent_id => p_ent_id
                   ) loop
      l_funent_exists := true;
      exit;
   end loop funent;

   bllog.write
   ('Create Function Attribute Usage if it does not already exist.'
   , bllog.debug
   );

   if l_funent_exists
   then
      <<funatt>>
      for r_funatt in c_funatt
                      ( b_fun_id => p_fun_id
                      , b_ent_id => p_ent_id
                      , b_att_id => p_att_id
                      ) loop
         l_funatt_exists := true;
         exit;
      end loop funatt;
   end if;   -- l_funent_exists

   if not l_funatt_exists
   then
      if not l_funent_exists
      then
         l_funent_id := blfunent.ins
                        ( function_reference  => p_fun_id
                        , entity_reference    => p_ent_id
                        , retrieve_flag       => 'Y'
                        );
      end if; -- not l_funent_exists

      l_funatt_id := blfunatt.ins
                     ( function_reference  => p_fun_id
                     , entity_reference    => p_ent_id
                     , attribute_reference => p_att_id
                     , retrieve_flag       => 'Y'
                     );

      bllog.write
      ( 'Added attribute usage '
        ||att_name(p_att_id)
        ||' for Business Rule Function '
        ||hsu_name.get_name_and_path(p_fun_id)
        ||'.'
      , bllog.information
      );
   end if; -- not l_funatt_exists

end create_att_usage;


procedure create_rel_usage
( p_fun_id       number
, p_from_ent_id number
, p_to_ent_id   number
, p_rel_name    varchar2
) is
--
-- Purpose  This procedure creates relationship usages for the given function if it does
--          not already exist
--
-- Usage    p_fun_id      : the function id of the function to whcih the relationship
--                          usage should be added
--          p_from_ent_id : id of the entity from which the relationship is directed
--          p_to_ent_id   : id of the entity to which the relationship is directed
--          p_rel_name    : the name of the relationship
--
-- Remarks  -
--

   -- checks whether there is an existing entity usage
   cursor c_funent
   (      b_fun_id in number
   ,      b_ent_id in number
   )      is
   select fent.id
   ,      fent.remark
   from   ci_function_entity_usages  fent
   where  fent.function_reference  = b_fun_id
   and    fent.entity_reference    = b_ent_id
   ;

   l_funent_exists boolean := false;
   l_funent_id     ci_function_entity_usages.id%type;
   l_remark        ci_function_entity_usages.remark%type := null;
begin

   bllog.write
   ('Create Function Entity Usage with relationship usage if it does not already exist.'
   , bllog.debug
   );

   <<funent>>
   for r_funent in c_funent
                   ( b_fun_id => p_fun_id
                   , b_ent_id => p_from_ent_id
                   ) loop
      l_funent_exists := true;
      l_remark        := r_funent.remark;
      l_funent_id     := r_funent.id;
      exit;
   end loop funent;

   if not l_funent_exists
   then
      l_funent_id := blfunent.ins
                     ( function_reference  => p_fun_id
                     , entity_reference    => p_from_ent_id
                     , retrieve_flag       => 'Y'
                     , remark              => '<RL>'
                                              ||p_rel_name
                                              ||'</RL>'
                     );
      bllog.write
      ( 'Relationship usage '
        ||hsu_name.get_name_and_path(p_from_ent_id)
        ||' '
        ||p_rel_name
        ||'added for Business Rule Function '
        ||hsu_name.get_name_and_path(p_fun_id)
        ||'.'
      , bllog.information
      );
   elsif l_remark is not null
   then
      if instr ( replace
                 ( upper(l_remark)
                 , ' '
                 , ''
                 )
               , '<RL>'
                 ||replace
                   (upper(p_rel_name)
                    , ' '
                    , ''
                    )
               , 1
               ) = 0
      then
         blfunent.upd
         ( p_funent_id => l_funent_id
         , remark      => l_remark
                          ||' <RL>'
                          ||p_rel_name
                          ||'</RL>'
         );

         bllog.write
         ( 'Relationship usage '
           ||hsu_name.get_name_and_path(p_from_ent_id)
           ||' '
           ||p_rel_name
           ||' added for Business Rule Function '
           ||hsu_name.get_name_and_path(p_fun_id)
           ||'.'
         , bllog.information
         );
      else
         bllog.write
         ( 'Relationship usage '
           ||hsu_name.get_name_and_path(p_from_ent_id)
           ||' '
           ||p_rel_name
           ||' already existent for Business Rule Function '
           ||hsu_name.get_name_and_path(p_fun_id)
           ||'.'
         , bllog.debug
         );
      end if; -- instr(replace...
   else
      <<funent2>>
      for r_funent in c_funent
                      ( b_fun_id => p_fun_id
                      , b_ent_id => p_from_ent_id
                      ) loop
         l_funent_id := r_funent.id;
      end loop funent2;

      blfunent.upd
      ( p_funent_id => l_funent_id
      , remark      => '<RL>'
                       ||p_rel_name
                       ||'</RL>'
      );

      bllog.write
      ( 'Added relationship usage '
        ||hsu_name.get_name_and_path(p_from_ent_id)
        ||' '
        ||p_rel_name
        ||' for Business Rule Function '
        ||hsu_name.get_name_and_path(p_fun_id)
        ||'.'
      , bllog.information
      );
   end if; -- not l_funent_exists

end create_rel_usage;


procedure create_modfun
( p_fun_id     in number
, p_fun_label  in varchar2
, p_plm_id     in number
, p_plm_name   in varchar2
) is
--
-- Purpose  This procedure adds module function usage if it does not already exist
--
-- Usage    p_fun_id    : id of the function the usage is added to
--          p_fun_label : the label of this function
--          p_plm_id    : id of the module the usages should be added to
--          p_plm_name  : the name of the plsql module
--
-- Remarks  -
--

   -- checks whether there is an existing module function link
   cursor c_modfun
   (      b_fun_id in number
   ,      b_mod_id in number
   )      is
   select 'x'
   from   ci_module_functions  modf
   where  modf.function_reference  = b_fun_id
   and    modf.module_reference    = b_mod_id
   ;

   l_modfun_id     ci_module_functions.id%type;
   l_modfun_exists boolean := false;
begin

   bllog.write
   ('Create Reference between module and function if it does not already exist.'
   , bllog.debug
   );

   <<modfun>>
   for r_modfun in c_modfun
                   ( b_fun_id => p_fun_id
                   , b_mod_id => p_plm_id
                   ) loop
      l_modfun_exists := true;
      exit;
   end loop modfun;

   if not l_modfun_exists
   then
      l_modfun_id := blmodfun.ins
                     ( function_reference  => p_fun_id
                     , module_reference    => p_plm_id
                     );
      bllog.write
      ( 'Module Function usage is created between PL/SQL Definition '
        ||hsu_name.get_name_and_path(p_plm_id)
        ||' and Business Rule Function '
        ||hsu_name.get_name_and_path(p_fun_id)
        ||'.'
      , bllog.information
      );
   end if; -- not l_modfun_exists
end create_modfun;


function determine_rule_type
( p_fun_label  in varchar2
, p_create_cco in varchar2
) return       varchar2
--
-- Purpose  This function determines the type of rule based on the given function label.
--          The result is a one character letter, which is either:
--            * S for a Static rule
--            * D for a Dynamic rule
--            * W for a change event rule with DML
--            * O for a change event rule without DML
--            * C for a Check Constraint
--            * X when it cannot be determined what kind of rule it is
--
-- Usage    p_fun_label : is the function label of the business rule function for
--                        which the type should be determined
--
-- Remarks  -
--
is
   l_rule varchar2(3);
begin
   bllog.write
   ( 'Determine rule type for '
     ||p_fun_label
     ||'.'
   , bllog.debug
   );
   g_transition_rule := false;
   g_delete_rule := false;
   l_rule := substr
             ( p_fun_label
             , length(p_fun_label) - 2
             , 3
             );
   if l_rule in ( 'ATT'
                , 'TPL'
                )
   then
      -- 6.5.3.4 use p_create_cco
      if p_create_cco = 'Y'
      then
         bllog.write
         ( p_fun_label
           ||' will become check constraint.'
         , bllog.debug
         );
         return 'C';
      else
         bllog.write
         ( p_fun_label
           ||' will become static br rule.'
         , bllog.debug
         );
         return 'S';
      end if;
   elsif l_rule in ( 'ENT'
                   , 'RER'
                   , 'IER'
                   )
   then
      bllog.write
      ( p_fun_label
        ||' is static.'
      , bllog.debug
      );
      return 'S';
   elsif l_rule in ( 'CRE'
                   , 'TRS'
                   , 'UPD'
                   , 'MOD'
                   , 'RDL'
                   , 'DEL'
                   )
   then
      -- remember if it is an attribute transition or delete rule
      if l_rule = 'TRS'
      then
         g_transition_rule := true;
      elsif l_rule = 'DEL'
      then
        g_delete_rule := true;
      end if; -- l_rule
      bllog.write
      ( p_fun_label
        ||' is dynamic.'
      , bllog.debug
      );
      return 'D';
   elsif l_rule in ( 'DFT'
                   , 'CEV'
                   )
   then
      bllog.write
      ( p_fun_label
        ||' is change event with DML.'
      , bllog.debug
      );
      return 'W';
   elsif l_rule = 'CEW'
   then
      bllog.write
      ( p_fun_label
        ||' is change event without DML.'
      , bllog.debug
      );
      return 'O';
   elsif l_rule = 'AUT'
   then
      bllog.write
      ( p_fun_label
        ||' is an atomic Authorization rule. This rule needs to be decomposed before '
        ||'it can be transformed.'
      , bllog.warnings
      );
      return 'X';
   else
      bllog.write
      ( p_fun_label
        ||' is not following CDM Standard OMS-60005. The rule cannot be transformed.'
      , bllog.warnings
      );
      return 'X';
   end if;

end determine_rule_type;


function rel_found
( p_remark    in varchar2
, p_rel_name  in varchar2
, p_to_ent    in number
)
return boolean
--
-- Purpose  This function checks whether a relationship can be determined from
--          the remark in an entity usage, and returns TRUE of this is the situation.
--          If not the function returns FALSE.
--
-- Usage    p_remark   : the remark wherein the relationship is searched.
--          p_rel_name : the name of the relationship that is searched
--          p_to_ent   : the name of the entity to which the relationship is directed
--
-- Remarks  -
--
is
   r_ent            blent.c_ent%rowtype;
   l_find_pos       number(3);
   l_other_end_ent  varchar2(240);
   l_ent_length     number(3);
begin
   -- 6.5.1.0  get name of 'to' entity
   open blent.c_ent(p_to_ent);
   fetch blent.c_ent into r_ent;
   close blent.c_ent;
   l_other_end_ent := replace(upper(r_ent.ent_name), ' ', '');
   l_find_pos := instr
                  ( replace
                    ( upper(p_remark)
                    , ' '
                    , ''
                    )
                  , '<RL>'
                    ||replace
                      ( upper(p_rel_name)
                      , ' '
                      , ''
                      )
                    ||'->'
                    ||l_other_end_ent
                    ||'</RL>'
                  , 1
                  );
   if l_find_pos != 0
   then
      bllog.write
      ('Relationship: '
       ||p_rel_name
       ||' is found to entity: '
       ||r_ent.ent_name
      , bllog.debug
      );
      return true;
   else
      bllog.write
      ('Relationship: '
       ||p_rel_name||'->'||r_ent.ent_name
       ||' is not found to be part of '
       ||p_remark
      , bllog.debug
      );
      return false;
   end if; -- l_find_pos != 0
end rel_found;

function get_brdd_name
( p_fun_label            in varchar2
, p_tbl_id               in number
)
return varchar2
is
   r_tbl                 bltbl.c_tbl%rowtype;
   l_decom_pos           number;
begin
    if g_trg_tbl.count = 1
    then
       return p_fun_label;
    else
       -- the rule is decomposed by the transformer
       l_decom_pos := instr
                      ( p_fun_label
                      , '_'
                      , 1
                      , 2
                      );
       -- get the table alias
       open bltbl.c_tbl(p_tbl_id);
       fetch bltbl.c_tbl into r_tbl;
       close bltbl.c_tbl;
       -- put table alias at position of decomposition id
       return substr( p_fun_label
                    , 1
                    , l_decom_pos - 1
                    )
              ||r_tbl.tbl_alias
              ||substr( p_fun_label
                      , l_decom_pos
                      )
              ;
    end if; -- g_trg_tbl.count
end get_brdd_name;

function transform_usages_to_args
( p_fun_id               in number
, p_fun_label            in varchar2
, p_tbl_id               in number
)
return varchar2
-- Purpose   returns the list of columns that will be arguments of the BRDD
--           based on the function entity (att/rel) usages for entities
--           implemented by the given table
is
   -- see if this entity is implemented by this table
   cursor c_tabent
   (      b_ent_id in number
   ,      b_tbl_id in number
   )      is
   select 1
   from   ci_tables_entities  tabent
   where  tabent.entity_reference = b_ent_id
   and    tabent.table_reference = b_tbl_id;

   l_dummy               number(1);
   l_col_list            varchar2(4000) := null;
   l_process_this_funatt boolean;
   l_process_this_funent boolean;
   l_col_id              ci_columns.id%type;
begin
   bllog.write
   ('Checking attribute retrieve usages for business rule '
    ||p_fun_label
    ||' and table '
    ||hsu_name.get_name_and_path(p_tbl_id)
   , bllog.debug
   );
   -- find out which of the function attribute usages correspond to this table
   <<funatt>>
   for r_funatt in blfunatt.c_funatt_ob_fun(p_fun_id)
   loop
      -- see if the entity referenced by this funatt is implemented by p_tbl_id
      open c_tabent(r_funatt.funatt_entity_reference, p_tbl_id);
      fetch c_tabent into l_dummy;
      l_process_this_funatt := c_tabent%found;
      close c_tabent;

      if l_process_this_funatt
      and r_funatt.funatt_retrieve_flag = 'Y'
      then
        l_col_id := find_col
                    ( p_tbl_id => p_tbl_id
                    , p_att_id => r_funatt.funatt_attribute_reference
                    );
        bllog.write
        ('Retrieve usage found for attribute: '
         ||att_name(r_funatt.funatt_attribute_reference)
         ||'; Column is: '
         ||to_char(l_col_id)
        , bllog.debug
        );
        l_col_list := l_col_list
                      || to_char(l_col_id)
                      || '|'
                      ;
      end if; -- l_process_this_funatt and retrieve = Y
   end loop funatt;

   -- check whether any foreign key columns are needed for validation
   -- this is when an entity is included as usage for the business rule
   -- function, with the relationship name in the comment
   bllog.write
   ('Checking relationship usages for business rule '
    ||p_fun_label
    ||' and table '
    ||hsu_name.get_name_and_path(p_tbl_id)
   , bllog.debug
   );
   -- find out which of the function attribute usages correspond to this table
   <<funent>>
   for r_funent in blfunent.c_funent_ob_fun(p_fun_id)
   loop
      -- see if the entity referenced by this funent is implemented by p_tbl_id
      open c_tabent(r_funent.funent_entity_reference, p_tbl_id);
      fetch c_tabent into l_dummy;
      l_process_this_funent := c_tabent%found;
      close c_tabent;

      if l_process_this_funent
      and r_funent.funent_retrieve_flag = 'Y'
      and instr(r_funent.funent_remark, '<RL>') > 0
      then
         bllog.write
         ('Retrieve usage with RL tag is found for entity '
          ||r_funent.ent_name
         , bllog.debug
         );

         -- loop over relationships starting in this entity
         <<relend>>
         for r_relend in blrelend.c_relend_ent(r_funent.funent_entity_reference)
         loop
            if rel_found
               ( p_remark   => r_funent.funent_remark
               , p_rel_name => r_relend.relend_name
               , p_to_ent   => r_relend.relend_to_entity_reference
               )
            then
               bllog.write
               ('Relationship: '
                 ||r_relend.relend_name
                 ||' is included as Usage for entity: '
                 ||r_funent.ent_name
               , bllog.debug
               );

               <<rel_columns>>
               for r_fco_col in c_fco_col
                                ( b_relend_id => r_relend.relend_id
                                , b_tbl_id    => p_tbl_id
                                )
               loop
                  l_col_list := l_col_list
                                || to_char(r_fco_col.column_reference)
                                || '|'
                                ;
               end loop rel_columns;
            end if; -- rel_found
         end loop relend;
      end if; -- process this funent
   	-- 6.5.3.4AMIS : added to have Argument created for (surrogate) Primary Key Column not defined as attribute
	 bllog.write
        ('Check whether FUNENT Remark contains '||PRIMARY_KEY_TAG||'. If so, create Arguments for PK Columns. Remark= '
         ||r_funent.funent_remark
        , bllog.debug
        );

       if instr(r_funent.funent_remark,PRIMARY_KEY_TAG) > 0 -- the tag <PK/> was included in the Comment for the Function Eneity Usage
       then
       	for pk_use in c_pk_use( b_fun_id => p_fun_id, b_ent_id => r_funent.funent_entity_reference
                      , b_tbl_id => p_tbl_id
					  ) loop
                 l_col_list := l_col_list
                               || to_char( pk_use.col_id)
                               || '|'
                               ;
                  bllog.write
                        ('An argument will be created'
                        ||' for primary key column '||pk_use.col_name
                        , bllog.debug
                        );
					  
               end loop;										  
       end if; -- Primary Key tag in FUNENT.remark
       -- end 6.5.4.3AMIS
   end loop funent;

   return l_col_list;
end transform_usages_to_args;


-- added in 6.5.3.4AMIS
procedure delete_brdds
( p_tbl_id in ci_table_definitions.id%type
, p_br_name in varchar2
)
-- Purpose  Delete existing Business Rule Design Definition
--
-- Usage    from run procedure
--
-- Remarks  Find PL/SQl Module Definition based on Business Rule Label
--
is
   cursor   c_brdd
   ( b_tbl_id in ci_table_definitions.id%type
   , b_br_name in varchar2
   )
   is
   select   trg.id                          trg_id
   ,        trg.name                        name
   ,        plm.id                          plm_id
   from     ci_database_triggers trg
   ,        ci_plsql_modules     plm
   where    trg.table_definition_reference = b_tbl_id
   and      trg.plsql_module_reference     = plm.id
   and      plm.name                       = b_br_name
   ;
begin
   -- loop over journalling busrules of this application
   <<brdd>>
   for r_brdd in c_brdd(p_tbl_id, p_br_name) loop
      -- delete busrule (delete of trigger is since 6i not enough anymore, so
      -- explicitly delete PL/SQL module also)
      bllog.write
      ( 'Deleting trigger '||r_brdd.name||' and PL/SQL module '
        ||hsu_name.get_name_and_path(r_brdd.plm_id)
        ||', a new business rule design definition overwrites this old one.'
      , bllog.information
      );
      bltrg.del(r_brdd.trg_id);
      blplm.del(r_brdd.plm_id);
   end loop brdd;
end delete_brdds;
---- end 6.5.3.4AMIS


procedure create_brdd_or_cc
-- create business rule design definition or check constraint, depending on the type
( p_rule_type             in varchar2
, p_tbl_id                in number
, p_brdd_name             in varchar2
, p_fun_id                in number
, p_fun_label             in varchar2
, p_fun_short_definition  in varchar2
, p_action                in varchar2
, p_triggering_columns    in varchar2
, p_when_condition        in varchar2
, p_col_data_required     in varchar2
, p_old_col_data_required in varchar2
, p_message_prefix        in varchar2
, p_message_number        in varchar2
, p_message_language      in varchar2
, p_overwrite_br         in varchar2 default 'N' -- 6.5.3.4AMIS
)
is
   r_tbl                 bltbl.c_tbl%rowtype;
   l_tbl_alias           ci_table_definitions.alias%type := null;
   l_already_exists      boolean;
   l_cco_name            ci_check_constraints.name%type;
   l_cco_id              ci_check_constraints.irid%type;
   l_trg_id              ci_database_triggers.id%type;
   l_plm_id              ci_plsql_modules.id%type;
begin
   -- 6.5.3.3 if rule type = C, create check constraint instead of BRDD
   if p_rule_type = 'C'
   then
      -- get the table alias
      open bltbl.c_tbl(p_tbl_id);
      fetch bltbl.c_tbl into r_tbl;
      close bltbl.c_tbl;
      l_tbl_alias := r_tbl.tbl_alias;
      -- derive check constraint name
      l_cco_name := hsu_name.get_next_cco_name
                    ( p_table_id      => p_tbl_id
                    , p_table_alias   => l_tbl_alias
                    , p_logical_name  => p_brdd_name
                    );
      --
      -- check if a Check Constraint with this name already exists
      l_already_exists := hsu_crbr.find_cco
                          ( p_cco_name => l_cco_name           -- in
                          , p_tbl_id   => p_tbl_id             -- in
                          , p_cco_id   => l_cco_id             -- out
                          );
      if l_already_exists
      then
         -- check constraint with this name already exists,
         -- business rule is skipped
         bllog.write
         ('A Check Constraint called '
          ||l_cco_name
          ||' already exists for table '
          ||hsu_name.get_name_and_path(p_tbl_id)
          ||'. This rule is skipped to prevent overwriting.'
          , bllog.warnings
          );
      end if; -- l_already_exists
   else -- p_rule_type <> 'C'
      -- 6.5.3.4AMIS
      if p_overwrite_br = 'Y'
	    then
        delete_brdds
        ( p_tbl_id => p_tbl_id
        , p_br_name => p_brdd_name 
        ); 
  	  end if; -- find_trg and overwrite =Y
      -- end 6.5.3.4AMIS
      -- check if BRDD with this plm_name already exists
      l_already_exists := hsu_capi.find_plm                   -- 6.5.0.1  extra parameters added
                          ( p_plm_short_name => p_brdd_name          -- in
                          , p_tbl_id         => p_tbl_id             -- in
                          , p_type           => 'BRDD'               -- in
                          , p_plm_id         => l_plm_id             -- out
                          );
      if l_already_exists
      then
         -- BRDD with this name already exists,
         -- business rule is skipped
         bllog.write
         ('A Business Rule Design Definition called '
          ||p_brdd_name
          ||' already exists. This rule is skipped to prevent overwriting.'
          , bllog.warnings
         );
      end if; -- BRDD exists
   end if; -- rule type = C

   if not l_already_exists
   then
      hsu_crbr.maintain_brdd
      ( p_utility_package       => PACKAGE_NAME
      , p_utility_revision      => revision
      , p_br_name               => p_brdd_name
      , p_description           => p_fun_short_definition
      , p_tbl_id                => p_tbl_id
      , p_tbl_alias             => l_tbl_alias
      , p_rule_type             => p_rule_type
      , p_action                => p_action
      , p_triggering_columns    => p_triggering_columns
      , p_when_condition        => p_when_condition
      , p_col_data_required     => p_col_data_required
      , p_old_col_data_required => p_old_col_data_required
      , p_message_prefix        => p_message_prefix
      , p_message_number        => p_message_number
      , p_message_language      => p_message_language
      , p_trg_id                => l_trg_id
      , p_plm_id                => l_plm_id
      , p_cco_id                => l_cco_id
      );
      if p_rule_type <> 'C'
      then
         -- create Module Function usage for the created pl/sql module and the business
         -- rule function
         create_modfun
         ( p_fun_id    => p_fun_id
         , p_fun_label => p_fun_label
         , p_plm_id    => l_plm_id
         , p_plm_name  => p_brdd_name
         );
      end if; -- p_rule_type
   end if; -- l_already_exists
end create_brdd_or_cc;


procedure transform_stage2
( p_fun_id               in number
, p_fun_label            in varchar2
, p_fun_short_definition in varchar2
, p_msg_prefix           in varchar2
, p_msg_language         in varchar2
, p_rule_type            in varchar2
, p_overwrite_br         in varchar2 default 'N' -- 6.5.3.4AMIS
) is
--
-- Purpose  This procedure transforms the given business rule function to design objects
--
-- Usage    p_fun_id               : is the function id of the function that should be
--                                   transformed
--          p_fun_label            : the label of the given function
--          p_fun_short_definition : the short definition of the function
--          p_msg_prefix           : the message prefix that should be used for the
--                                   business rule
--          p_msg_language         : the language used for the message
--
-- Remarks  -
--

   -- l_triggering_columns is a list of column ids that when updated trigger business
   -- rule enforcement; only relevant when p_action contains 'U'
   l_triggering_columns  varchar2(5000);
   l_arg_columns         varchar2(5000);
   l_old_columns         varchar2(5000);
   l_action              varchar2(10); -- any combination of I(nsert), U(pdate) and D(elete)
   l_trg_id              ci_database_triggers.id%type;
   l_plm_id              ci_plsql_modules.id%type;
   l_cco_id              ci_check_constraints.id%type;
   l_brdd_name           ci_modules.name%type;
   l_msg_prefix          varchar2(15);
   l_msg_number_c        varchar2(15);
begin
   bllog.write( 'Stage 2: loop over all tables found and create BRDD.'
              , bllog.debug
              );
   <<triggering_tables>>
   for i in 1..g_trg_tbl.count loop
      bllog.write
      ('...Create BRDD for table '
      ||hsu_name.get_name_and_path(g_trg_tbl(i).tbl_id)
      ||'.'
      , bllog.debug
      );

      -- create list of triggering columns for this table
      l_triggering_columns:= '';
      <<trg_columns>>
      for j in 1..g_trg_col.count loop
         if g_trg_col(j).tbl_id = g_trg_tbl(i).tbl_id
         then
            l_triggering_columns:= l_triggering_columns
                                   ||to_char(g_trg_col(j).col_id)
                                   ||'|';
         end if;
      end loop trg_columns;

      -- determine the list of columns that will be arguments
      -- based on the function entity usages for entities implemented by this table
      l_arg_columns := transform_usages_to_args( p_fun_id
                                               , p_fun_label
                                               , g_trg_tbl(i).tbl_id
                                               );

      -- determine the name of the BRDD
      l_brdd_name := get_brdd_name(p_fun_label, g_trg_tbl(i).tbl_id);

      -- 6.5.1.0  moved action stuff to after the plm name is determined
      -- so it can be used in the warning
      l_action:='';
      if g_trg_tbl(i).ins
      then
         l_action:= l_action||'I|';
      end if;
      if g_trg_tbl(i).upd
      then
         l_action:= l_action||'U|';
      end if;
      if g_trg_tbl(i).del
      then
         l_action:= l_action||'D|';
      end if;
      if l_action = ''
      then
         -- 6.5.1.0  fix for bug 1566730: no ins, no upd, no del: give warning
         bllog.write
         ( 'Unable to determine any triggering event (insert, update, delete)'
           ||' for Business Rule '||l_brdd_name
           ||' on table '||hsu_name.get_name_and_path(g_trg_tbl(i).tbl_id)
           ||', correct this before creating the CAPI.'
         , bllog.warnings
         );
      end if;

      if g_trg_tbl(i).when_condition is not null
      then
         -- when condition was created because this is a supertype table
         if p_rule_type = 'C' -- 6.5.3.3
         then
            bllog.write
            ( 'Business Rule '||l_brdd_name||' is triggered by a Subtype Entity,'
              ||' so an extra condition is added to the Check Constraint'
              ||' Where/Validation Condition to ensure that the rule only fires for this Subtype.'
            );
            bllog.write
            ( 'It is assumed that the discriminator column of table '
              ||hsu_name.get_name_and_path(g_trg_tbl(i).tbl_id)
              ||' will not be updateable.'
              ||' If it is updateable, you might have to manually change the'
              ||' Where/Validation Condition of the Check Constraint for '||l_brdd_name||'.'
            , bllog.warnings
            );
         else
            bllog.write
            ( 'Business Rule '||l_brdd_name||' is triggered by a Subtype Entity,'
              ||' so a When Condition is added to the Triggering Events'
              ||' to ensure that the rule only fires for this Subtype.'
            );
            bllog.write
            ( 'It is assumed that the discriminator column of table '
              ||hsu_name.get_name_and_path(g_trg_tbl(i).tbl_id)
              ||' will not be updateable.'
              ||' If it is updateable, you might have to manually change the trigger part of'
              ||' Business Rule Design Definition '||l_brdd_name||'.'
            , bllog.warnings
            );
         end if; -- p_ruletype
      end if; -- when condition

      if p_rule_type in ('W', 'O')
      then
         -- change event, use dummy message
         l_msg_prefix := DUMMY_MSG_PREFIX;
         l_msg_number_c := DUMMY_MSG_NUMBER;
      else
         l_msg_prefix := p_msg_prefix;
         l_msg_number_c := null;
      end if; -- p_rule_type

      -- check if old column values must be included
      if g_transition_rule
      then
         l_old_columns := l_arg_columns;
      elsif g_delete_rule
      then
         l_old_columns := l_arg_columns;
         l_arg_columns := null;
      else
         l_old_columns := null;
      end if; -- g_transition_rule

      -- 6.5.1.3  before checking for existing PLM, set the correct context folder
      -- if default context folder, set it to folder of table
      if blapi.get_default_context_fol_flag
      then
         blapi.set_context_fol
         ( p_context_folder_irid => jr_acc_rights.get_owning_container(g_trg_tbl(i).tbl_id)
         );
      end if;

      -- create business rule design definition or check constraint,
      -- depending on the rule type
      create_brdd_or_cc
      ( p_rule_type             => p_rule_type
      , p_tbl_id                => g_trg_tbl(i).tbl_id
      , p_brdd_name             => l_brdd_name
      , p_fun_id                => p_fun_id
      , p_fun_label             => p_fun_label
      , p_fun_short_definition  => p_fun_short_definition
      , p_action                => l_action
      , p_triggering_columns    => l_triggering_columns
      , p_when_condition        => g_trg_tbl(i).when_condition
      , p_col_data_required     => l_arg_columns
      , p_old_col_data_required => l_old_columns
      , p_message_prefix        => l_msg_prefix
      , p_message_number        => l_msg_number_c
      , p_message_language      => p_msg_language
      -- 6.5.3.4AMIS
      , p_overwrite_br          => p_overwrite_br
      );

   end loop triggering_tables;
end; -- transform_stage2


procedure process_subtype
( p_ent_id       in ci_entities.id%type
, p_tbl_id       in ci_table_definitions.id%type
, p_discr_col_id in ci_columns.id%type
, p_discr_value  in ci_tables_entities.discriminator_value%type
, p_on_condition in ci_events.on_condition%type
, p_rule_type    in varchar2
)
-- Purpose  Add correct when condition to business rule design definition
--          because business rule might be for subtype only within supertype table
--
is
   r_ent             blent.c_ent%rowtype;
   l_discr_col_id    ci_columns.id%type;
   l_discr_col_name  ci_columns.name%type;
   l_when_condition  varchar2(1000);
begin
   -- when this procedure is called we know that it concerns a supertype implementation
   -- but we don't know if this entity is the supertype or a subtype
   open blent.c_ent(p_ent_id);
   fetch blent.c_ent into r_ent;
   close blent.c_ent;
   if r_ent.ent_supertype_reference is not null
   then
      -- it is a subtype, we need a when condition
      l_discr_col_id := nvl
                       ( p_discr_col_id
                       , find_discriminator_col_id(p_tbl_id)
                       );
      l_discr_col_name := col_name(l_discr_col_id);
      if p_rule_type = 'C' -- 6.5.3.3
      then
         l_when_condition := l_discr_col_name
                             ||' = '''
                             ||p_discr_value||''''
                          ;
      else
         l_when_condition := 'nvl(:new.'
                             ||l_discr_col_name
                             ||', :old.'
                             ||l_discr_col_name
                             ||') = '''
                             ||p_discr_value||''''
                          ;
      end if; -- p_rule_type
      -- add when condition to the triggering from this table
      add_tbl
      ( p_tbl_id    => p_tbl_id
      , p_event     => p_on_condition
      , p_when      => l_when_condition
      );
   end if; -- supertype_reference
end process_subtype;

function transform_funevt
( p_fun_row           in blfun.c_fun%rowtype
, p_funevt_row        in blfunevt.c_funevt_oe_fun%rowtype
, p_create_att_usages in varchar2
, p_rule_type         in varchar2
)
return boolean -- can br be transformed?
is
   l_non_owned_relend       boolean;
   l_ent_id                 ci_entities.id%type;
   r_tabent                 bltabent.c_tabent_oe_ent%rowtype;
   l_tabent_found           boolean;
   l_relend_id              ci_relationship_ends.id%type;
   l_to_ent_id              ci_entities.id%type;
   l_other_relend_id        ci_relationship_ends.id%type;
   l_superent_id            ci_entities.id%type;
begin
   bllog.write
   ('Triggering event '||p_funevt_row.evt_name||' found.'
   , bllog.debug);

   l_non_owned_relend := false;
   l_ent_id           := p_funevt_row.evt_entity_reference;
   -- Check whether the event refers to updating a relationship
   if  p_funevt_row.evt_on_condition = ON_CONDITION_UPDATE
   and p_funevt_row.evt_attribute_reference is null
   then
      -- if it is a relationship event, get the relend properties
      check_relend_event
      ( p_funevt_row.evt_id
      , p_funevt_row.evt_name
      , l_ent_id        -- in / out
      , l_relend_id
      , l_to_ent_id
      , l_other_relend_id
      , l_superent_id
      , l_non_owned_relend
      );
   end if; -- check for relationship event
   if l_ent_id is null and not l_non_owned_relend then
      bllog.write
      ( 'Relationship end ignored, no corresponding Foreign Key found.'
      , bllog.warnings
      );
   else -- FK found for relationship
      -- find (first) table_entity mapping for entity
      -- if none, check if supertype: perhaps sub-type implementation in separate tables??
      l_tabent_found:= false;
      open bltabent.c_tabent_oe_ent(l_ent_id);
      fetch bltabent.c_tabent_oe_ent into r_tabent;
      l_tabent_found := bltabent.c_tabent_oe_ent%found;
      close bltabent.c_tabent_oe_ent;
      if l_tabent_found
      then
         -- if found, check discrimininator column and value
         -- (COLUMN_REFERENCE, DISCRIMINATOR_VALUE)
         if r_tabent.tabent_discriminator_value is not null
         then
            -- a supertype implementation
            process_subtype
            ( p_ent_id       => l_ent_id
            , p_tbl_id       => r_tabent.tabent_table_reference
            , p_discr_col_id => r_tabent.tabent_column_reference
            , p_discr_value  => r_tabent.tabent_discriminator_value
            , p_on_condition => p_funevt_row.evt_on_condition
            , p_rule_type    => p_rule_type
            );
         end if; -- supertype implementation

         if l_non_owned_relend
         then
            bllog.write
            ( 'Relationship End is not owned (l_non_owned_relend).'
            , bllog.debug
            );

            add_tbl
            ( p_tbl_id => r_tabent.tabent_table_reference
            , p_event  => 'INSERT'
            );
            add_tbl
            ( p_tbl_id => r_tabent.tabent_table_reference
            , p_event  => 'DELETE'
            );
            add_tbl
            ( p_tbl_id => r_tabent.tabent_table_reference
            , p_event  => 'UPDATE'
            );

            -- loop over all columns in the foreign key implementing l_other_relend_id
            for r_fco_col in c_fco_col
                             ( b_relend_id => l_other_relend_id
                             , b_tbl_id    => r_tabent.tabent_table_reference
                             ) loop
               add_trg_col
               ( p_tbl_id => r_tabent.tabent_table_reference
               , p_col_id => r_fco_col.column_reference
               );
            end loop; -- r_fco_col
         else
            add_tbl
            ( p_tbl_id => r_tabent.tabent_table_reference
            , p_event  => p_funevt_row.evt_on_condition
            );

            if l_relend_id is not null
            then
               bllog.write
               ( 'Relationship End is owned (l_non_owned_relend=FALSE).'
                 ||'relend_id:'
                 ||to_char(l_relend_id)
                 ||' tbl_id:'
                 ||to_char(r_tabent.tabent_table_reference)
               , bllog.debug
               );

               -- loop over all columns in the foreign key implementing l_relend_id
               for r_fco_col in c_fco_col
                                ( b_relend_id => l_relend_id
                                , b_tbl_id    => r_tabent.tabent_table_reference
                                ) loop
                    add_trg_col
                    ( p_tbl_id => r_tabent.tabent_table_reference
                    , p_col_id => r_fco_col.column_reference
                    );

                  bllog.write
                  ( 'Added FK triggering column with id: '
                    ||r_fco_col.column_reference
                  , bllog.debug_detailed
                  );
               end loop; -- r_fco_col

               -- if usages should be entered for the business rule
               -- function from the triggering events, then usages should
               -- be created for the relationship
               if p_create_att_usages = 'Y'
               then
                  bllog.write
                  ( 'Relationship Usages should be created '
                  ||'(if they don''t exist alteready).'
                  , bllog.debug
                  );
                  create_rel_usage
                  ( p_fun_id      => p_fun_row.fun_id
                  , p_from_ent_id => l_ent_id
                  , p_to_ent_id   => l_to_ent_id
                  , p_rel_name    => relend_name(l_relend_id)
                  );
               end if; --p_create_att_usage = 'Y''

            elsif p_funevt_row.evt_attribute_reference is not null
            then
               -- add p_trg_col the column implementing the attribute
               add_trg_col
               ( p_tbl_id => r_tabent.tabent_table_reference
               , p_col_id => find_col
                             ( p_tbl_id => r_tabent.tabent_table_reference
                             , p_att_id => p_funevt_row.evt_attribute_reference)
               );
               -- if attribute usages should be entered for the business rule
               -- function from the triggering attributes, then this should
               -- occur here
               if p_create_att_usages = 'Y'
               then
                  create_att_usage
                  ( p_fun_id    => p_fun_row.fun_id
                  , p_ent_id    => l_ent_id
                  , p_att_id    => p_funevt_row.evt_attribute_reference
                  );
               end if;
            end if; -- l_relend_id is not null
         end if; -- l_non_owned_relend
      else -- not l_tabent_found
         bllog.write
         ( 'No table exists for triggering entity '
           ||hsu_name.get_name_and_path(l_ent_id)
           ||' of triggering event '||p_funevt_row.evt_name
           ||'. Business Rule '||p_fun_row.fun_function_label||' is not transformed.'
         , bllog.warnings
         );
         return false;
      end if;
   end if; -- 6.5.2.2
   return true;
end transform_funevt;

function transform_funent_as_funevt
( p_fun_row           in blfun.c_fun%rowtype
, p_funent_row        in c_funent%rowtype
, p_rule_type         in varchar2
)
return boolean -- can br be transformed?
is

   -- selects entity and attribute usages for the given business rule function
   cursor c_fun_ent_att
   (      b_fun_id   in number
   ,      b_ent_id   in number
   ) is
   select fatt.attribute_reference
   from   ci_function_attribute_usages  fatt
   where  fatt.function_reference        = b_fun_id
   and    fatt.entity_reference         = b_ent_id
   ;

   -- selects relationship ends required for business rule validation
   cursor   c_relend_use
   (        b_fun_id       in number
   ,        b_ent_id        in number
   ,        b_superent_id  in number
   )        is
   select   relend.id
   ,        relend.name
   ,        relend.to_entity_reference
   ,        funent.remark
   from     ci_function_entity_usages  funent
   ,        ci_relationship_ends       relend
   where    funent.function_reference       =  b_fun_id
   and      funent.entity_reference         =  b_ent_id
   and      ( relend.from_entity_reference    =  funent.entity_reference
              or
              -- required when subtype, and relationship to supertype
              ( b_superent_id is not null
                and
                relend.from_entity_reference  =  b_superent_id
              )
              -- required when subtype, and relationship to subtype
              or
              ( b_superent_id is not null
                and
                relend.from_entity_reference  =  b_ent_id
              )
            )
   and      instr(funent.remark, '<RL>') != 0
   ;

   l_superent_id            ci_entities.id%type;
   l_previous_ent_id        ci_entities.id%type;
   r_tabent                 bltabent.c_tabent_oe_ent%rowtype;
   l_tabent_found           boolean;
   l_tab_id                 ci_relation_definitions.id%type;
begin
   l_superent_id := null;
   -- find table_entity for entity
   -- if none, check if supertype: perhaps sub-type implementation in separate tables??
   -- l_tabent_found:= false;
   if l_previous_ent_id != p_funent_row.entity_reference
   or l_previous_ent_id is null
   then
      l_previous_ent_id := p_funent_row.entity_reference;
      l_tabent_found := false;
      -- take the first table that implements this entity
      open bltabent.c_tabent_oe_ent(p_funent_row.entity_reference);
      fetch bltabent.c_tabent_oe_ent into r_tabent;
      l_tabent_found := bltabent.c_tabent_oe_ent%found;
      close bltabent.c_tabent_oe_ent;
      if l_tabent_found
      then
         -- check if it is supertype implementation
         if r_tabent.tabent_discriminator_value is not null
         then
            -- supertype implementation
            -- (we assume insert and update triggers,
            -- so we can use update to add the when condition)
            process_subtype
            ( p_ent_id       => p_funent_row.entity_reference
            , p_tbl_id       => r_tabent.tabent_table_reference
            , p_discr_col_id => r_tabent.tabent_column_reference
            , p_discr_value  => r_tabent.tabent_discriminator_value
            , p_on_condition => 'UPDATE'
            , p_rule_type    => p_rule_type
            );

            <<super_ent>>
            for c_ent in blent.c_ent(p_funent_row.entity_reference) loop
               l_superent_id := c_ent.ent_supertype_reference;
            end loop super_ent;
         end if; -- supertype implementation

         add_tbl
         ( p_tbl_id => r_tabent.tabent_table_reference
         , p_event  => 'INSERT'
         );
         add_tbl
         ( p_tbl_id => r_tabent.tabent_table_reference
         , p_event  => 'UPDATE'
         );
         bllog.write
         ('Triggers will be created for Business Rule '
          ||p_fun_row.fun_function_label
          ||' on table belonging to entity: '
          ||hsu_name.get_name_and_path(p_funent_row.entity_reference)
         , bllog.debug
         );
         -- remember the table for later use
         l_tab_id := r_tabent.tabent_table_reference;
      else -- not l_tabent_found
         bllog.write
         ( 'No table exists for entity usage '||hsu_name.get_name_and_path(p_funent_row.entity_reference)
           ||'. Business Rule '||p_fun_row.fun_function_label
           ||' is not transformed.'
         , bllog.warnings
         );
         return false;
      end if; -- l_tabent_found
   end if;  -- previous ent id != current

   bllog.write
   ('Search function attribute usages for '
    ||p_fun_row.fun_function_label
    ||' with id: '
    ||p_fun_row.fun_id
    ||' and entity '
    ||hsu_name.get_name_and_path(p_funent_row.entity_reference)
    ||' with id: '
    ||p_funent_row.entity_reference
   , bllog.debug
   );

   <<fun_ent_att>>
   for r_fun_ent_att in c_fun_ent_att
                        ( b_fun_id => p_fun_row.fun_id
                        , b_ent_id => p_funent_row.entity_reference
                        )
   loop
      add_trg_col
      ( p_tbl_id => l_tab_id
      , p_col_id => find_col
                    ( p_tbl_id => l_tab_id
                    , p_att_id => r_fun_ent_att.attribute_reference
                    )
      );
      bllog.write
      ('Triggering columns and arguments will be created'
       ||' from Business Rule '
       ||p_fun_row.fun_function_label
       ||' for column belonging to attribute: '
       ||att_name(r_fun_ent_att.attribute_reference)
      , bllog.debug
      );
   end loop fun_ent_att;

   bllog.write
   ('Search relationship usages for '
    ||p_fun_row.fun_function_label
    ||' with id: '
    ||p_fun_row.fun_id
    ||' and entity '
    ||hsu_name.get_name_and_path(p_funent_row.entity_reference)
    ||' with id: '
    ||p_funent_row.entity_reference
   , bllog.debug
   );

   <<relend>>
   for r_relend_use in c_relend_use
                       ( b_fun_id      => p_fun_row.fun_id
                       , b_ent_id      => p_funent_row.entity_reference
                       , b_superent_id => l_superent_id
                       )
   loop
      if r_relend_use.id is not null
      then
         if rel_found
            ( p_remark   => r_relend_use.remark
            , p_rel_name => r_relend_use.name
            , p_to_ent   => r_relend_use.to_entity_reference
            )
         then
            bllog.write
            ('Relationship (foreign key column): '
             ||r_relend_use.name
             ||' is included as argument and as triggering column.'
             , bllog.debug
            );

            <<rel_columns2>>
            for r_fco_col in c_fco_col
                             ( b_relend_id => r_relend_use.id
                             , b_tbl_id    => l_tab_id
                             )
            loop
               add_trg_col
               ( p_tbl_id        => l_tab_id
               , p_col_id        => r_fco_col.column_reference
               );
            end loop rel_columns2;
         end if; -- rel_found
      end if; -- r_relend_use.id is not null
   end loop relend;
   return true;
end transform_funent_as_funevt;

function transform_stage1
( p_fun_row           in blfun.c_fun%rowtype
, p_rule_type         in varchar2
, p_msg_prefix        in varchar2
, p_msg_language      in varchar2
, p_create_att_usages in varchar2
)
return boolean -- can rule be transformed?
is
   l_event_found            boolean;
   l_usage_found            boolean;
   l_table_found            boolean;

begin
   bllog.write
   ( 'Stage 1: Populate in memory table with all triggering tables and columns.'
   , bllog.debug
   );

   -- for the business rule, find all entities that trigger events for it
   g_trg_tbl.delete;
   g_trg_col.delete;
   l_event_found := false;
   l_table_found := false;

   <<trg_events>>
   for r_funevt in blfunevt.c_funevt_oe_fun(p_fun_row.fun_id)
   loop
      -- triggering event valid?
      if  r_funevt.evt_entity_reference is null
      then
         bllog.write
         ('Triggering event '
          ||r_funevt.evt_name
          ||' is invalid, it has no Entity reference.'
          -- 6.5.2.1
          -- ||' Business Rule '||r_fun.fun_function_label||' has not been transformed.'
          ||' This triggering event is ignored.'
          -- 6.5.2.1
         , bllog.warnings
         );
      else
         l_event_found := true;
         l_table_found := transform_funevt
                          ( p_fun_row
                          , r_funevt
                          , p_create_att_usages
                          , p_rule_type
                          );
         if not l_table_found
         then
            exit; -- br will not be transformed
         end if;
      end if; -- entity reference is null
   end loop trg_events;

   if l_event_found
   then
      if l_table_found
      then
         bllog.write('l_event_found=true, l_table_found=true', bllog.DEBUG_DETAILED);
      else
         bllog.write('l_event_found=true, l_table_found=false', bllog.DEBUG_DETAILED);
      end if;
   else
      bllog.write('l_event_found=false', bllog.DEBUG_DETAILED);
   end if;

   if l_event_found and not l_table_found -- 6.5.3.3
   then -- if for one of the triggering events no table is found,
        -- then the rule is not transformed (already reported)
      return false;
   else
     -- If there are no triggering events and the rule is not static
     -- then the rule cannot be transformed.
     -- If the rule is static, then the function entity/attribute usages are checked.
     if not l_event_found
     then
        if p_rule_type in ('C', 'S') -- 6.5.3.3 also for check constraints
        then
           bllog.write
           ('Search function entity usages for '
            ||p_fun_row.fun_function_label
            ||' with id: '
            ||p_fun_row.fun_id
           , bllog.debug
           );

           l_usage_found := false;
           <<funent>>
           for r_funent in c_funent
                             ( b_fun_id => p_fun_row.fun_id
                           )
           loop
              l_usage_found := true;
              l_table_found := transform_funent_as_funevt
                               ( p_fun_row
                               , r_funent
                               , p_rule_type
                               );
              if not l_table_found
              then
                 exit; -- br can not be transformed
              end if;
           end loop funent;
           if l_usage_found and not l_table_found -- 6.5.3.3
           then
              return false;
           elsif not l_usage_found -- 6.5.3.3
           then
              bllog.write
                 ('The Static Business Rule Function '
                  ||p_fun_row.fun_function_label
                  ||' does not have triggering events, nor retrieve usages.'
                  ||'  Rule cannot be transformed.'
                  , bllog.warnings
                  );
              return false;
           end if; -- l_usage_found / l_table_found
        else -- rule type is not static or check constraint
           bllog.write
           ('Business Rule Function '
            ||p_fun_row.fun_function_label
            ||' does not have triggering events.  Rule cannot be transformed.'
            , bllog.warnings
            );
            return false;
        end if; -- l_rule_type = 'S' or 'C'
     end if; -- not l_event_found

     return true;
   end if;
end transform_stage1;

procedure transform_br_fun
( p_fun_id            in number
, p_msg_prefix        in varchar2
, p_msg_language      in varchar2
, p_create_att_usages in varchar2
, p_create_cco        in varchar2
, p_overwrite_br      in varchar2 default 'N' -- 6.5.3.4AMIS
) is
--
-- Purpose  This procedure finds all triggering events that trigger Business Rule
--          Function p_fun_id.
--            It will call procedure create_brdd for each table implementing one or more
--          triggering entities to create one Business Rule Design Definition.
--
-- Usage    -
--
-- Remarks  -
--
   l_rule_type              varchar2(1) := 'X'; -- ruletype not determined;
begin
   <<fun>>
   for r_fun in blfun.c_fun(p_fun_id) loop
      bllog.write
      ('<BR><B>Transformation of Business Rule '
       ||hsu_name.get_name_and_path(r_fun.fun_irid)   -- 6.5.0.0  path added
       ||'.</B>'
      , bllog.information
      );

      -- first populate the tables g_trg_tbl and g_trg_col with all triggering tables
      -- and columns.
      -- Then looping over all the tables found
      --    check non-existence of BRDD in table
      --    make call to create_brdd to actually create the BR Design Definition
      --    create MODFUN between PLM created and BRFUN


      -- check whether the function is atomic; if it is not, do not process
      -- also don't process common functions
      if is_atomic(r_fun.fun_irid) = 'N' -- 6.5.0.0  via function, atomic flag removed in 6i
      then
         bllog.write
         ('Business Function '
          ||r_fun.fun_function_label
          ||' cannot be transformed since it is not atomic'
          ||' (it has child-functions). '
         , bllog.warnings
         );
      elsif r_fun.fun_common_function_reference is not null
      then
         bllog.write
         ('Business Function '
          ||r_fun.fun_function_label
          ||' cannot be transformed since it is a common function. '
         , bllog.warnings
         );
      else
         -- for each atomic, non-common business rule function
         -- determine type of business rule.
         -- This may be one of Static (S), Dynamic (D),
         -- Change event with DML (W), change event without DML (O),
         -- or Check Constraint (C).
         l_rule_type := determine_rule_type
                        ( p_fun_label  => r_fun.fun_function_label
                        , p_create_cco => p_create_cco
                        );
         if not l_rule_type = 'X' -- rule can not be transformed
         then
            if transform_stage1
               ( r_fun
               , l_rule_type
               , p_msg_prefix
               , p_msg_language
               , p_create_att_usages
               )
            then
               transform_stage2
               ( p_fun_id               => r_fun.fun_id
               , p_fun_label            => r_fun.fun_function_label
               , p_fun_short_definition => r_fun.fun_short_definition
               , p_msg_prefix           => p_msg_prefix
               , p_msg_language         => p_msg_language
               , p_rule_type            => l_rule_type
               , p_overwrite_br         => p_overwrite_br -- 6.5.3.4AMIS
               );
            end if; -- stage 1 succeeded
         end if; -- l_rule_type = 'X'
      end if; -- not function_atomic or common
   end loop fun;
end; -- transform_br_fun

procedure run
( p_brfun_list        in varchar2 default null
, p_fun_name_filter   in varchar2 default null
, p_root_fun_id       in number   default null
, p_ent_list          in varchar2 default null
, p_create_att_usages in varchar2
, p_create_cco        in varchar2                     -- 6.5.3.3  new
, p_msg_prefix        in varchar2 default '<default>' -- 6.5.0.0  changed
, p_msg_language      in varchar2
, p_overwrite_br      in varchar2 default'N' -- 6.5.3.4AMIS
)
--
-- Purpose  Runs the utility
--
-- Usage    p_brfun_list is a multi select list of functions that the utility
--                       should process
--          p_fun_name_filter is a name filter (inclusing wildcards % and _); all
--                       Business Rule Functions whose name passes the filter condition
--                       will be processed
--          p_root_fun_id is the id of one Business Rule Function; if it has a value,
--                       then the utilityw ill process the entire function-tree starting
--                       at p_root_fun_id and including all its sub-functions
--          p_ent_list is a multi select list of entities; if this parameter has a value,
--                       then for all indicated entities the utility will find all
--                       Business Rule Functions they trigger and all those Business
--                       Rule Functions will the utility process
--          p_create_att_usages indicates whether there should be created attribute
--                       usages for the business rule function from the attributes given
--                       in the triggering event
--          p_create_cco indicates whether check constraints should be created for
--                       ATT and TPL rules
--          p_msg_prefix gives the prefix that should be used for messages
--          p_msg_languate indicates the langueage in which the message should be written.
--          p_overwrite_br indicates whether an existing business rule may be overwritten
--
-- Remarks  Only one of the first 4 parameters needs to have a value.
--          The first one that actually has a value determines the utility's behavior,
--          regardless of values of subsequent parameters.
is
   -- t_fun_list holds id's of all business rule functions that must be processed
   t_fun_list     hsu_par.list_items_table_type;
   -- t_ent_list holds id's of all entities for which the business rules they trigger
   -- must be processed
   t_ent_list     hsu_par.list_items_table_type;
   l_fun_tbl      t_num_tbl;
   l_par_count    number(1);
   l_br_fun_found boolean;

   cursor c_fun
   ( b_filter in varchar2
   ) is
   select   fun.id
   from     ci_functions fun
   where    hsu_brtr.is_atomic(fun.irid)    =    'Y' -- 6.5.0.0  atomic property via function
   and      fun.function_label              like upper(b_filter)
   order by fun.function_label
   ;

   cursor c_fun_tree
   (p_root_fun in number
   ) is
   select fun.id
   from   ci_functions fun
   where  hsu_brtr.is_atomic(fun.irid) = 'Y'
   start with fun.id                   = p_root_fun
   connect by prior fun.id             = fun.parent_reference
   ;

   cursor c_entfun
   ( b_ent_id in number
   ) is
   select distinct
          fun.id
   from   ci_events evt
   ,      ci_event_triggers evttrg
   ,      ci_functions fun
   where  evt.entity_reference                 =    b_ent_id
   and    evttrg.triggered_event_reference     =    evt.id
   and    evttrg.triggering_function_reference =    fun.id
   and    hsu_name.is_business_rule(fun.function_label) = 'Y' -- 6.5.0.0   check br label
   ;                                                          -- via function

begin
   -- Check whether only a single parameter value is given.  Otherwise an error is raised.
   l_par_count := 0;
   if p_brfun_list is not null
   then
      l_par_count := l_par_count + 1;
   end if; -- p_brfun_list is not null

   if p_fun_name_filter is not null
   then
      l_par_count := l_par_count + 1;
   end if; -- p_fun_name_filter is not null

   if p_root_fun_id is not null
   then
      l_par_count := l_par_count + 1;
   end if; -- p_root_fun_id is not null

   if p_ent_list is not null
   then
      l_par_count := l_par_count + 1;
   end if; -- p_ent_list is not null

   if l_par_count > 1
   then
      -- more parameters are given, so raise an error
      bllog.write
      ('Only one of the following parameters are allowed:'
      ||'<BR> - Business Rule function'
      ||'<BR> - Filter for BR Functions'
      ||'<BR> - BR Function Tree Root'
      ||'<BR> - Rules triggered by Entity'
      , bllog.errors
      );
      raise e_user_error;
   elsif l_par_count = 0
   then
      -- none of the parameters are given, so raise an error
      bllog.write
      ('One of the following parameters must be supplied:'
      ||'<BR> - Business Rule function'
      ||'<BR> - Filter for BR Functions'
      ||'<BR> - BR Function Tree Root'
      ||'<BR> - Rules triggered by Entity'
      , bllog.errors
      );
      raise e_user_error;
   else
      -- start running transformer
      if p_brfun_list is not null
      then
         t_fun_list:= hsu_par.get_list_items
                      ( p_brfun_list
                      , '|'
                      );
         <<fun_list>>
         for i in 1 .. t_fun_list.last loop
            transform_br_fun
            ( p_fun_id            => t_fun_list(i)
            , p_msg_prefix        => p_msg_prefix
            , p_msg_language      => p_msg_language
            , p_create_att_usages => p_create_att_usages
            , p_create_cco        => p_create_cco
         		, p_overwrite_br         => p_overwrite_br -- 6.5.3.4AMIS
            );
         end loop fun_list;

      elsif p_fun_name_filter is not null
      then
         <<fun_filter>>
         for r_fun in c_fun
                     (b_filter => p_fun_name_filter
                     ) loop
            transform_br_fun
            ( p_fun_id            => r_fun.id
            , p_msg_prefix        => p_msg_prefix
            , p_msg_language      => p_msg_language
            , p_create_att_usages => p_create_att_usages
            , p_create_cco        => p_create_cco
         		, p_overwrite_br         => p_overwrite_br -- 6.5.3.4AMIS
            );
         end loop fun_filter;

      elsif p_root_fun_id is not null
      then
         <<fun_tree>>
         for r_fun_tree in c_fun_tree
                          (p_root_fun => p_root_fun_id
                          ) loop
             transform_br_fun
            ( p_fun_id            => r_fun_tree.id
            , p_msg_prefix        => p_msg_prefix
            , p_msg_language      => p_msg_language
            , p_create_att_usages => p_create_att_usages
            , p_create_cco        => p_create_cco
         		, p_overwrite_br         => p_overwrite_br -- 6.5.3.4AMIS
            );
         end loop fun_tree;

      elsif p_ent_list is not null
      then
         t_ent_list:= hsu_par.get_list_items
                      ( p_ent_list
                      , '|'
                      );
         <<ent_list>>
         for i in 1..t_ent_list.last loop
            l_br_fun_found := false;
            <<ent_fun>>
            for r_entfun in c_entfun
                           ( b_ent_id => to_number(t_ent_list(i))
                           ) loop
               l_br_fun_found := true;
               if not exists_in_num_tbl
                      ( p_num     => r_entfun.id
                      , p_num_tbl => l_fun_tbl
                      )
               then
                  add_to_num_tbl
                  ( p_num     => r_entfun.id
                  , p_num_tbl => l_fun_tbl
                  );
               end if; -- not exists_in_num_tbl
            end loop ent_fun;
         end loop ent_list;
         if not l_br_fun_found
         then
            bllog.write
            ('No business rules found for the specified entities. '
            ||hsu_name.br_notification
            , bllog.errors
            );
            raise e_user_error;
         end if; -- l_br_fun_found

         <<fun_tbl>>
         for i in 1 .. l_fun_tbl.last loop
             transform_br_fun
            ( p_fun_id            => l_fun_tbl(i)
            , p_msg_prefix        => p_msg_prefix
            , p_msg_language      => p_msg_language
            , p_create_att_usages => p_create_att_usages
            , p_create_cco        => p_create_cco
         		, p_overwrite_br         => p_overwrite_br -- 6.5.3.4AMIS
            );
         end loop fun_tbl;

     else
       bllog.write('Tranform Business Rule Functions to Business Rule Design elements '
                 ||' could not run. None of the parameters has a value!'
               , bllog.WARNINGS);
     end if; -- p_all_entities ='Y'

  end if; --l_par_count > 1
  bllog.write('Completed hsu_BRTR.run', bllog.DEBUG);
exception
   when others then
      if blapi.handle_exception then
         raise;
      end if;
end run;

end hsu_brtr;
/
