create or replace package hsu_BRTR
is

function is_atomic
( p_fun_irid in ci_functions.irid%type
)
return varchar2;
--
-- Purpose  returns Y if function has no child functions, N otherwise
--
-- Usage    from hsu_brtr.install (parameter where clause)
--
-- Remarks  -
--
pragma restrict_references(is_atomic, WNDS, WNPS);


function revision
return varchar2;
--
-- Purpose  returns the revision label of the utility
--
-- Usage    -
--
-- Remarks  -
--
pragma restrict_references(revision, WNDS, RNDS, WNPS);

procedure install;
--
-- Purpose  installs the utility into the Headstart Utilities tables
--
-- Usage    -
--
-- Remarks  -
--

procedure deinstall;
--
-- Purpose  removes the utility from the Headstart Utilities tables.
--
-- Usage    -
--
-- Remarks  -
--

procedure run
( p_brfun_list        in varchar2 default null
, p_fun_name_filter   in varchar2 default null
, p_root_fun_id       in number   default null
, p_ent_list          in varchar2 default null
, p_create_att_usages in varchar2
, p_create_cco        in varchar2                     -- 6.5.3.3  new
, p_msg_prefix        in varchar2 default '<default>' -- 6.5.0.0  changed
, p_msg_language      in varchar2
, p_overwrite_br      in varchar2 default'N' -- 6.5.3.4AMIS
);
--
-- Purpose  Runs the utility
--
-- Usage    p_brfun_list is a multi select list of functions that the utility
--                       should process
--          p_fun_name_filter is a name filter (inclusing wildcards % and _); all
--                       Business Rule Functions whose name passes the filter condition
--                       will be processed
--          p_root_fun_id is the id of one Business Rule Function; if it has a value,
--                       then the utilityw ill process the entire function-tree starting
--                       at p_root_fun_id and including all its sub-functions
--          p_ent_list is a multi select list of entities; if this parameter has a value,
--                       then for all indicated entities the utility will find all
--                       Business Rule Functions they trigger and all those Business
--                       Rule Functions will the utility process
--          p_create_att_usages indicates whether there should be created attribute
--                       usages for the business rule function from the attributes given
--                       in the triggering event
--          p_create_cco indicates whether check constraints should be created for
--                       ATT and TPL rules
--          p_msg_prefix gives the prefix that should be used for messages
--          p_msg_languate indicates the langueage in which the message should be written.
--          p_overwrite_br indicates whether an existing business rule may be overwritten
--
-- Remarks  Only one of the first 4 parameters needs to have a value.
--          The first one that actually has a value determines the utility's behavior,
--          regardless of values of subsequent parameters.

end hsu_brtr;
/
