create or replace package body hsu_crbr
/*****************************************************************************************
Purpose         (Re)Create a Business Rule in a Designer trigger and PL/SQL definition

Usage           To create input for utility Generate CAPI

Remarks         -

Revision History  (most recent on top!)
When          Who                                Construct
  Revision    What
------------------------------------------------------------------------------------------
06-aug-2005   Lucas Jellema (AMIS)
   6.5.3.3AMIS   Added code to have names of available arguments added to the Pl/SQL Block
05-apr-2004   Marc Vahsen                        maintain_cco, create_message_if_needed,
  6.5.3.3                                        maintain_brdd
              Message constraint_name naming convention for check constraints will differ
              from BR design definitions, as the check constraint name already contains
              application alias.
              Remove <p1> parameter for br's for check constraints
25-feb-2004   Sandra Muller                      maintain_brdd, maintain_cco, find_cco,
                                                 add_where_condition, install
  6.5.3.2     Bugfix 2761012: Add support for creating business rules as check constraints
16-jan-2002   Sandra Muller
  6.5.2.0     fix for Oracle 9i: same default value for p_msg_prefix in spec and body
21-dec-2000   Sandra Muller
  6.5.1.0     (upper) br_name must be prefixed with appl.code!
              see revision history for 6.0.0.12
21-dec-2000   Sandra Muller
  6.5.1.0     In initial checks, searching message with constraint should be done with prefix before br name
16-nov-2000   Sandra Muller
  6.5.0.2     Changed references to RuleFrame information
20-oct-2000   Marc Vahsen
  6.5.0.1     Change in parameter list of hsu_capi.find_plm
19-oct-2000   Marc Vahsen
  6.5.0.0     Upgrade to designer 6i
              - handle default context folder
              - check message format for new messages codes
              - handle <default> message prefix
24-mar-2000   Sandra Muller                      maintain_brdd
  6.0.0.20    Removed call to blapi.handle_exception from exception handler, see bug 1247704
20-mar-2000   Sandra Muller                      default_code (constant), install
  6.0.0.19    Changed comment text and added l_rule_ok :=
              + changed help text of validation code parameter
13-mar-2000   Lauri L. Boyd                      default_code (constant)
  6.0.0.18    Added trace statement to default code.
13-mar-2000   Sandra Muller                      install
  6.0.0.17    Changed help text for arguments whose prompts where changed in 6.0.0.16
10-mar-2000   Sandra Muller                      maintain_brdd, add_plsql_block, install
  6.0.0.16    If other utilities call maintain_brdd, use their package name and revision
              in the Revision History of the BRDD
              + new prompts for Cols Required For BR + Validation code param before Msg pars
07-mar-2000   Sandra Muller                      add
  6.0.0.15    Split up line if longer than 2000 characters
06-mar-2000   Sandra Muller                      install
  6.0.0.14    Changed help text when condition: use :old/:new instead of p_old/p_new
03-mar-2000   Sandra Muller                      install
  6.0.0.13    Removed references to Ctrl+E from help text
01-mar-2000   Sandra Muller                      create_message_if_needed
  6.0.0.12    Prefix br name with app code in message constraint name
18-feb-2000   Sandra Muller                      initial_checks
  6.0.0.11    Check other message already exists for this rule only if msg number not null
17-feb-2000   Sandra Muller                      maintain_brdd, create_message_if_needed
  6.0.0.10    Turned create_message_if_needed into function + moved forward
              otherwise plm does not get right message code
16-feb-2000   Sandra Muller                      run, install, maintain_brdd, maintain_plm
  6.0.0.9     Split up message code parameter in msg prefix and number, derive number if
              not specified and insert message
10-feb-2000   Sandra Muller                      maintain_trg
  6.0.0.8     Made trigger an After Statement trigger, documentation only (not used by CAPI)
07-feb-2000   Sandra Muller                      install, run, maintain_brdd, add_notes,
                                                 initial_checks, maintain_plm, maintain_trg
  6.0.0.7     Made p_description mandatory as it is also used for message text.
              + put rule type + message code in Notes instead of Purpose
              + put description in Purpose (both trg and plm)
              + extra check if message for this business rule exists with other code
              + use new function hil_message.msg_code_exists
04-feb-2000   Sandra Muller                      DEFAULT_CODE
  6.0.0.6     bugfix: forgot ; after l_rule_ok declaration
04-feb-2000   Sandra Muller                      initial_checks, install, maintain_brdd
  6.0.0.5     create default message if it does not exist, extra language parameter
              + name change
03-feb-2000   Sandra Muller                      maintain_trg, maintain_brdd
  6.0.0.4     Added trigger description
03-feb-2000   Sandra Muller                      run, maintain_brdd, install, initial_checks
  6.0.0.4                                        , maintain_plm, add_plsql_block
              Changes because Validation Code will already include l_rule_ok declaration,
              return and default exception handler, replacing [br_name]
31-jan-2000   Sandra Muller                      maintain_trg
  6.0.0.4     use hsu_capi.find_trg
28-jan-2000   Sandra Muller                      maintain_trg + initial_checks
  6.0.0.3     maintain_trg: fixed bug of value error when triggered on Update
                            but no triggering columns
              initial_checks: use hsu_capi.chk_msg_code_exists
24-jan-2000   Sandra Muller                      add_plsql_block
  6.0.0.2     Use bltext.add_revision_history, get_comment_start_line, get_comment_end_line
21-jan-2000   Sandra Muller                      initial_checks, maintain_plm
  6.0.0.1     Moved function chk_col_datatype to hsu_capi package + use hsu_capi.find_plm
21-jan-2000   Sandra Muller                      run, install, initial_checks
                                                 add, add_arguments, add_plsql_block
                                                 get_plm_purpose, maintain_plm
                                                 maintain_trg, maintain_brdd
  6.0.0.0     Create trigger and pl/sql definition instead of add to capi package definition
              includes new parameter When condition, run procedure completely rewritten
15-11-1999    Lauri L. Boyd                      run, install
  2.0         Changed input parms to allow for 4 rule types
                   - Static Data Constraint
                   - Dynamic Data Constraint
                   - Change Event with DML
                   - Change Event without DML
              Removed Enforcement Level, Time of Enforcement, and Static or Dynamic
              Changed to conform to pl/sql coding standards
              Changed order of input parms
              Added initial_checks procedure to validate input parms.
              Added exception if no capi found for requested table.
26-09-1999    Lucas Jellema                      run, install
  1.3         New parameter to indicate static/dynamic character of business rule
26-08-1999    Lucas Jellema                      run, install
  1.2         Changed Parameter Enforcement Level (better titles)
23-08-1999    Lucas Jellema                      run, install
  1.1         Create Business Rule, add parameter p_old_col_data_required
17-08-1999    Lucas Jellema
  1.0         Creation, based on HSU_VAPI; using HSU_CAPI
*****************************************************************************************/
/*
$add_to_revision_history
$
*/
is

--
-- private types
--

--
-- private constants
--
PACKAGE_NAME         constant varchar2(30)   := 'HSU_CRBR';
REVISION_LABEL       constant varchar2(30)   := '$x.y:: 6.5.3.3  $';
UTILITY_NAME         constant varchar2(60)   := 'Create Business Rule Design Definition';
-- max length br name is minimum of max length plm.name, plm.short_name, trg.name
-- and plm.implementation_name
MAX_LENGTH_BR_NAME    constant number(2,0)   := 20;
MAX_LENGTH_ARG_NAME   constant number(2,0)   := 30;
MAX_LENGTH_PURPOSE    constant number(2,0)   := 70;
MAX_LENGTH_CCO_REMARK constant number(2,0)   := 70;
MAX_LENGTH_CCO_NAME   constant number(2,0)   := 30;
-- following three depend on definition of qms message tables
MAX_LENGTH_MSG_CODE  constant number(2,0)    := 15;
MAX_LENGTH_MSG_DESCR constant number(3,0)    := 500;
MAX_LENGTH_MSG_TEXT  constant number(3,0)    := 200;
-- default value of validation code parameter
DEFAULT_CODE         constant varchar2(2000) :=
               '   l_rule_ok boolean := true;'
    ||chr(10)||'begin'
    ||chr(10)||'   trace(''[br_name] (f)'');'
    ||chr(10)
    ||chr(10)||'   -- for instructions, see the Headstart User Guide'
    ||chr(10)||'   l_rule_ok := '
    ||chr(10)
    ||chr(10)||'   return l_rule_ok;'
    ||chr(10)||'exception'
    ||chr(10)||'   when others'
    ||chr(10)||'   then'
    ||chr(10)||'      qms$errors.unhandled_exception(PACKAGE_NAME||''.[br_name] (f)'');'
    ||chr(10)||'end [br_name];'
   ;
FORMAT_RATIONALE_1 constant   varchar2(2000) :=
   'The message code does not comply with CDM standard OMG-10040, which states that '
   ||'a message should consist of 3 letters, a dash, and 5 numbers (for instance '
   ||'QMS-00100).'
   ;
FORMAT_RATIONALE_2 constant   varchar2(2000) :=
   'If you want to deviate from this standard, you will have to modify the public '
   ||'constants c_errorformat and possibly also c_errorlength (maybe even the function '
   ||'CheckErrorFormat) in package qms$errors, both on the server and (if you use the '
   ||'Template Package) in the Headstart core library.'
   ;
FORMAT_RATIONALE_3 constant   varchar2(2000) :=
   'Otherwise the message text will not be shown when the message is raised.'
   ||chr(10)
   ;

--
-- private variables
--

-- global package variables to build up a multi-line text block
g_text         bltext.cdi_text_lines;
g_text_counter number(5,0);
g_arguments    varchar2(1000); -- 6.5.3.3AMIS

--
-- private cursors
--

--
-- private exceptions
--
e_initial_checks          exception;
e_abort_utility           exception;

--
-- forward declarations
--

--
-- private program units
--

procedure add
( p_text_line    in     varchar2
)
-- Purpose  adds a single text string to the next available line in
--          pl/sql table g_text
--          if the text line is longer than 2000, split it at a chr(10)
--
is
   l_text_line varchar2(30000);
   l_nl_pos    number(5);
begin
   l_text_line := p_text_line;
   while length(l_text_line) > 2000 loop
      -- find the first chr(10) after the 1000th character
      l_nl_pos := instr(l_text_line, chr(10), 1000);
      if l_nl_pos = 0 or l_nl_pos > 2000
      then
         bllog.write
         ( 'The supplied text contains a piece of 1000 characters'
           ||' without any newline character (chr(10) in it.'
         , bllog.errors
         );
      else
         g_text_counter := g_text_counter + 1;
         g_text(g_text_counter) := substr(l_text_line, 1, l_nl_pos - 1);
         l_text_line := substr(l_text_line, l_nl_pos + 1);
      end if;
   end loop;
   -- write remaining piece of text line
   g_text_counter := g_text_counter + 1;
   g_text(g_text_counter) := l_text_line;
end add;

-- 6.5.3.3AMIS

procedure reset_arguments
is
begin
  g_arguments:='';
end;

procedure add_argument( p_argument_name in varchar2)
is
begin
  g_arguments := g_arguments||', '||p_argument_name;
end;    

function get_arguments
return varchar2
is
begin
  return g_arguments;
end;  
-- end 6.5.3.3AMIS

procedure add_arguments
( p_plm_id      in ci_plsql_modules.id%type
, p_column_list in varchar2
, p_name_prefix in varchar2
)
-- Purpose  Add Arguments to existing PL/SQL Definition,
--          with name <p_name_prefix>||p_col_name for all columns in p_column_list
is
   -- select maximum existing argument position for this module
   cursor c_arg
   (      b_plm_id in ci_plsql_modules.id%type
   )      is
   select nvl(max(arg.position), 0)
   from   ci_module_arguments arg
   where  arg.plsql_module_reference = b_plm_id;  -- 6.5.0.0  column name changed

   l_max_position        ci_module_arguments.position%type;
   l_col_table           hsu_par.list_items_table_type;
   r_col                 blcol.c_col%rowtype;
   l_arg_name            ci_module_arguments.name%type;
   l_arg_id              ci_module_arguments.id%type;
begin
   reset_arguments; -- 6.5.3.3AMIS

   -- determine maximum argument position
   open c_arg(p_plm_id);
   fetch c_arg into l_max_position;
   close c_arg;

   l_col_table := hsu_par.get_list_items(p_column_list, '|');
   <<col>>
   for i in 1 .. nvl(l_col_table.last, 0) loop
      open blcol.c_col(to_number(l_col_table(i)));
      fetch blcol.c_col into r_col;
      close blcol.c_col;
      l_arg_name := p_name_prefix
                    ||substr
                      ( lower(r_col.col_name)
                      , 1
                      , MAX_LENGTH_ARG_NAME - length(p_name_prefix)
                      );
      l_arg_id := blarg.ins
                  ( plsql_module_reference  => p_plm_id
                  , column_reference        => r_col.col_id
                  , input_output_identifier => 'INPUT'
                  , name                    => l_arg_name
                  , position                => l_max_position + i
                  );
      bllog.write
      ( 'Argument '||l_arg_name||' added for column '||r_col.col_name||'.'
      , bllog.information
      );
  	  add_argument(l_arg_name); -- 6.5.3.3AMIS
   end loop col;
end add_arguments;


procedure add_notes
( p_plm_id       in ci_plsql_modules.id%type
, p_rule_type    in varchar2
, p_message_code in varchar2
)
-- Purpose  Put rule type and message code in Notes of PL/SQL Definition as follows:
--          The following tags are used by the CAPI utility (don't remove!)
--          Add your own notes below these tags.
--          <TY>rule type</TY> -- the type of the business rule
--          <MG>message code</MG> -- the violation message code
--
-- Usage    Called from maintain_plm
--
-- Remarks
--
is
   l_line1   varchar2(70);
   l_line2   varchar2(70);
begin
   bltext.init_text_lines(g_text, g_text_counter);
   add('The following tags are used by the CAPI utility (don''t remove!)');
   add('Add your own notes below these tags.');
   -- rule type
   l_line1 := '<TY>';
   if p_rule_type = 'S'
   then
      l_line1 := l_line1||'Static Data Constraint';
   elsif p_rule_type = 'D'
   then
      l_line1 := l_line1||'Dynamic Data Constraint';
   elsif p_rule_type = 'W'
   then
      l_line1 := l_line1||'Change Event with DML';
   elsif p_rule_type = 'O'
   then
      l_line1 := l_line1||'Change Event without DML';
   end if;
   l_line1 := l_line1||'</TY> -- the type of the business rule';
   add(l_line1);

   -- message code
   l_line2 := '<MG>'||upper(p_message_code)||'</MG> -- the violation message code';
   add(l_line2);
   bllog.write
   ( 'Adding the following 2 lines to Notes of PL/SQL Definition: '
     ||'<BR>' -- html new line
     ||htf.escape_sc(l_line1)
     ||'<BR>' -- html new line
     ||htf.escape_sc(l_line2)
   , bllog.information
   );

   -- then call bltext.write_lines_to_repository
   bltext.write_lines_to_repository
   ( p_plm_id
   , 'PLM'
   , 'CDINOT'
   , g_text
   , g_text_counter
   , 'REPLACE'
   );
end add_notes;


procedure add_plsql_block
( p_plm_id             in ci_plsql_modules.id%type
, p_plm_already_exists in boolean
, p_description        in varchar2
, p_utility_package    in varchar2
, p_utility_revision   in varchar2
, p_validation_code    in varchar2
, p_br_name            in varchar2
, p_arguments          in varchar2 default '' -- 6.5.3.3AMIS
)
-- Purpose  Add or replace PL/SQL Block of PL/SQL Definition,
--          if p_plm_already_exists, preserve and add to Revision History
is
   l_new_revision_label varchar2(10);
begin
   -- put new text in variable of type cdi_text_lines
   bltext.init_text_lines(g_text, g_text_counter);

   -- write the comment block including (updated?) revision history
   add(bltext.get_comment_start_line);
   add('Purpose     '||p_description);
   add('Remarks     ');
   add('Input parameters available in the validation code: '||p_arguments); -- 6.5.3.3AMIS
   add('');
   l_new_revision_label := bltext.add_revision_history
                           ( g_text
                           , g_text_counter
                           , p_plm_already_exists
                           , p_plm_id
                           , p_utility_package
                           , p_utility_revision
                           );
   add(bltext.get_comment_end_line);
   -- after the comment block, write the whole validation code
   -- but replace [br_name] with the business rule name
   add(replace(p_validation_code, '[br_name]', lower(p_br_name)));

   -- then call bltext.write_lines_to_repository
   bltext.write_lines_to_repository
   ( p_plm_id
   , 'PLM'
   , 'CDIPLS'
   , g_text
   , g_text_counter
   , 'REPLACE'
   );
end add_plsql_block;


procedure add_where_condition
( p_cco_id             in ci_check_constraints.id%type
, p_when_condition     in varchar2
, p_validation_code    in varchar2
, p_br_name            in varchar2
)
-- Purpose  Add or replace Where/Validation Condition of Check Constraint
--
-- Usage    From maintain_cco
--
-- Remarks  New in 6.5.3.2
is
   l_replaced_when       varchar2(2000);
   l_replaced_validation varchar2(2000);
begin
   -- put new text in variable of type cdi_text_lines
   bltext.init_text_lines(g_text, g_text_counter);

   l_replaced_validation := replace(p_validation_code, '[br_name]', lower(p_br_name));

   if p_when_condition is null
   then
      add(l_replaced_validation);
   else
      l_replaced_when := replace(p_when_condition, ':new.', '');
      l_replaced_when := replace(l_replaced_when,  ':old.', '');
      -- add when condition to validation code, joined by 'not' and 'or'
      add('not ('||l_replaced_when||')');
      add('or');
      add('('||l_replaced_validation||')');
   end if; -- p_when_condition

   -- then call bltext.write_lines_to_repository
   bltext.write_lines_to_repository
   ( p_cco_id
   , 'CCO'
   , 'CDWHCL'
   , g_text
   , g_text_counter
   , 'REPLACE'
   );
end add_where_condition;


function find_cco
( p_cco_name  in   ci_check_constraints.name%type
, p_tbl_id    in   ci_table_definitions.id%type
, p_cco_id    out  ci_check_constraints.id%type
)
return boolean
-- Purpose  Check if this table already has a check constraint with this name,
--          and if it exists, provide the id of the check constraint in an out parameter
--
-- Usage    Just before you want to create such a check constraint
--
-- Remarks  New in 6.5.3.2
--
is
   l_cco_found boolean := false;
begin
   for r_cco in blcco.c_cco_tab(p_tbl_id) loop
      if r_cco.cco_name = p_cco_name
      then
         l_cco_found := true;
         p_cco_id := r_cco.cco_id;
         exit;
      end if;
   end loop;
   return l_cco_found;
end find_cco;


procedure initial_checks
( p_br_name               in varchar2
, p_rule_type             in varchar2
, p_action                in varchar2
, p_triggering_columns    in varchar2
, p_col_data_required     in varchar2
, p_old_col_data_required in varchar2
, p_message_prefix        in varchar2
, p_message_number_c      in varchar2
, p_context_app_code      in varchar2 -- 6.5.1.0  added
)
-- Purpose  validate input parameters, both warnings and errors
--
-- Usage    called from maintain_brdd
--
-- Remarks  if error(s), it raises an exception which should not be caught but
--          raised through to the utility start form
--
is
   l_check_failed          boolean                       := false;
   l_col_list              hsu_par.list_items_table_type;
   l_existing_msg_code     varchar2(20);
   l_message_code          varchar2(100);
begin
   -- warning if rule type is Check Constraint, and actions are not I and U
   if p_rule_type = 'C'
   then
      if instr(p_action, 'I') = 0
      or instr(p_action, 'U') = 0
      or instr(p_action, 'D') > 0
      then
         bllog.write
         ( 'Check Constraints always fire during both Insert and Update, and not during Delete.'
         , bllog.warnings
         );
      end if;
   end if; -- p_rule_type = 'C'
   -- warning if update columns specified while no update specified
   if  instr(p_action,'U') = 0
   and p_triggering_columns is not null
   then
      bllog.write
      ( 'Update Trigger Columns ignored because Update not included in '
        ||'Triggering DML Event.'
      , bllog.warnings
      );
   end if;

   -- warning if no update columns specified while update is specified
   if  instr(p_action,'U') > 0
   and p_triggering_columns is null
   then
      bllog.write
      ( 'No Update Trigger Columns specified.  Business Rule '
        ||p_br_name||' will fire when <I>any</I> column is updated.'
      , bllog.warnings
      );
   end if;

   -- error if p_br_name too long
   if length(p_br_name) > MAX_LENGTH_BR_NAME
   then
      l_check_failed := true;
      bllog.write
      ( 'Parameter Business Rule Name ('
        || p_br_name
        ||') is too long, maximum length is '
        || to_char(MAX_LENGTH_BR_NAME)
        ||'.'
      , bllog.errors
      );
   end if;

   -- error if p_br_name not alphanumeric with underscores
   if not hsu_name.check_alphanumeric(p_br_name, '_')
   then
      l_check_failed := true;
      bllog.write
      ( 'Parameter Business Rule Name ('
        || p_br_name
        ||') must be alphanumeric (underscores ''_'' are allowed).'
      , bllog.errors
      );
   end if;

   l_message_code := upper(p_message_prefix||'-'||p_message_number_c);
   -- error if other message with this br name already exists
   if p_message_number_c is not null
   and hil_message.msg_code_exists
      -- 6.5.1.0  (upper) br_name must be prefixed with appl.code!
      -- see revision history for 6.0.0.12
      ( p_constraint        => upper(p_context_app_code||'_'||p_br_name)
      , p_existing_msg_code => l_existing_msg_code
      )
   then
      if l_existing_msg_code <> l_message_code
      then
         bllog.write
         ( 'Other message '
           ||upper(l_existing_msg_code)
           ||' already exists for this Business Rule name.'
         , bllog.errors
         );
         l_check_failed := true;
      end if; -- l_existing_msg_code
   end if; -- msg_code_exists

   -- error if message prefix and number don't match the required message format
   -- 6.5.0.0  check message format via qms$errors.CheckErrorFormat
   if  p_message_number_c is not null
   then
      -- check the message code
      if not qms$errors.CheckErrorFormat(p_msgtxt => l_message_code)
      then
         bllog.write
         ( 'The message code '''||l_message_code||''' is not correct:'
           ||chr(10)
           ||FORMAT_RATIONALE_1
         , bllog.errors
         );
         bllog.write -- split up in 3 pieces (because of max length of text in bllog)
         ( FORMAT_RATIONALE_2
         , bllog.errors
         );
         bllog.write
         ( FORMAT_RATIONALE_3
         , bllog.errors
         );
         l_check_failed := true; -- 6.5.0.0  added
      end if;
   end if;

   -- error if wrong datatype of triggering columns,
   -- columns required for BR and old columns required for BR
   if p_triggering_columns is not null
   then
      l_col_list := hsu_par.get_list_items(p_triggering_columns||'|', '|');
      for i in 1..l_col_list.last loop
         for r_col in blcol.c_col(l_col_list(i)) loop
            if not hsu_capi.chk_col_datatype
                   ( r_col.col_datatype
                   , r_col.col_name
                   , 'Update Trigger column'
                   )
            then
               l_check_failed := true;
            end if;
         end loop; -- r_col in c_col
      end loop; -- i in 1..l_col_list.last
   end if; -- p_triggering_columns is not null

   if p_col_data_required is not null
   then
      l_col_list := hsu_par.get_list_items(p_col_data_required||'|','|');
      for i in 1..l_col_list.last loop
         for r_col in blcol.c_col(l_col_list(i)) loop
            if not hsu_capi.chk_col_datatype
                   ( r_col.col_datatype
                   , r_col.col_name
                   , 'Columns Required for BR, column'
                   )
            then
               l_check_failed := true;
            end if;
         end loop; -- r_col in c_col
      end loop; -- i in 1..l_col_list.last
   end if; -- if p_col_data_required is not null

   if p_old_col_data_required is not null
   then
      l_col_list := hsu_par.get_list_items(p_old_col_data_required,'|');
      for i in 1..l_col_list.last loop
         for r_col in blcol.c_col(l_col_list(i)) loop
            if not hsu_capi.chk_col_datatype
                   ( r_col.col_datatype
                   , r_col.col_name
                   , 'Old Columns Required for BR, column'
                   )
            then
               l_check_failed := true;
            end if;
         end loop; -- r_col in c_col
      end loop; -- i in 1..l_col_list.last
   end if; -- p_old_col_data_required

   if l_check_failed
   then
      raise e_initial_checks;
   end if;
   bllog.write('end of initial_checks', bllog.debug);
exception
   when others
   then
      raise;
end; -- initial_checks


function maintain_cco
( p_context_app_code      in  varchar2
, p_br_name               in  varchar2
, p_tbl_id                in  number
, p_tbl_alias             in  varchar2
, p_description           in  varchar2
, p_message_code          in  varchar2
, p_when_condition        in  varchar2
, p_validation_code       in  varchar2
)
-- Purpose  Maintains Check Constraint that represents Business Rule
--          and returns its id
--
-- Usage    Called from maintain_brdd
--
-- Remarks  New in 6.5.3.2
--
return ci_check_constraints.id%type
is
   l_cco_id              ci_check_constraints.id%type;
   l_cco_already_exists  boolean;
   l_cco_name            varchar2(150);
   l_cco_comment         ci_check_constraints.remark%type;
   l_tbl_alias           ci_table_definitions.alias%type;
   r_tbl                 bltbl.c_tbl%rowtype;
begin
   bllog.write('starting maintain_cco', bllog.debug);

   -- if table alias is not provided, derive it here
   if p_tbl_alias is null
   then
      -- get the table alias
      open bltbl.c_tbl(p_tbl_id);
      fetch bltbl.c_tbl into r_tbl;
      close bltbl.c_tbl;
      l_tbl_alias := r_tbl.tbl_alias;
   else
      l_tbl_alias := p_tbl_alias;
   end if; -- p_tbl_alias is null

   l_cco_name := hsu_name.get_next_cco_name
                 ( p_table_id      => p_tbl_id
                 , p_table_alias   => l_tbl_alias
                 , p_logical_name  => p_br_name
                 );

   if length(l_cco_name) > MAX_LENGTH_CCO_NAME
   then
      bllog.write
      ( 'Generated Check Constraint name "'||l_cco_name||'" is too long, it is abbreviated to "'
        ||substr(l_cco_name, 1, MAX_LENGTH_CCO_NAME)||'".'
      , bllog.warnings
      );
      l_cco_name := substr(l_cco_name, 1, MAX_LENGTH_CCO_NAME);
   end if; -- length cco_name

   l_cco_comment := substr
                    ( p_description
                    , 1
                    , MAX_LENGTH_CCO_REMARK
                    );

   -- check if a Check Constraint with this name already exists
   l_cco_already_exists := find_cco
                           ( p_cco_name => l_cco_name           -- in
                           , p_tbl_id   => p_tbl_id             -- in
                           , p_cco_id   => l_cco_id             -- out
                           );

   if not l_cco_already_exists
   then
      -- a new Check Constraint must be created
      l_cco_id := blcco.ins
                  ( name                          => l_cco_name
                  , table_reference               => p_tbl_id
                  , error_message                 => p_message_code
                  , check_constraint_type         => 'SQL EXPRESSION'
                  , remark                        => l_cco_comment
                  , complete_flag                 => 'Y'
                  , enabled_flag                  => 'Y'
                  , implementation_level          => 'SERVER'
                  );
      bllog.write
      ( 'Created Check Constraint: '
        ||hsu_name.get_name_and_path(p_tbl_id)
        ||'/'
        ||l_cco_name
        ||'.'
      , bllog.information
      );
   else
      -- recreate the existing Check Constraint
      -- update everything (except name, because that's how it was identified)
      blcco.upd
      ( p_cco_id                      => l_cco_id
      , table_reference               => p_tbl_id
      , error_message                 => p_message_code
      , check_constraint_type         => 'SQL EXPRESSION'
      , remark                        => l_cco_comment
      , complete_flag                 => 'Y'
      , enabled_flag                  => 'Y'
      , implementation_level          => 'SERVER'
      , defer_status                  => 'NOT DEFERRED'
      , exception_table               => null
      , client_enforce_when_null_flag => 'N'
      );
      bllog.write
      ( 'Recreated existing Check Constraint: '
        ||hsu_name.get_name_and_path(p_tbl_id)
        ||'/'
        ||l_cco_name
        ||'.'
      , bllog.information
      );
   end if; -- l_cco_already_exists

   add_where_condition
   ( l_cco_id
   , p_when_condition
   , p_validation_code
   , p_br_name
   );

   bllog.write('end of maintain_cco', bllog.debug);
   return l_cco_id;
end maintain_cco;



function maintain_plm
( p_br_name               in  varchar2
, p_tbl_id                in  number    -- 6.5.0.1  parameter added
, p_description           in  varchar2
, p_rule_type             in  varchar2
, p_col_data_required     in  varchar2
, p_old_col_data_required in  varchar2
, p_message_code          in  varchar2
, p_utility_package       in  varchar2
, p_utility_revision      in  varchar2
, p_validation_code       in  varchar2

)
-- Purpose  Maintains PL/SQL Definition part of Business Rule Design Definition
--          and returns its id
--
-- Usage    Called from maintain_brdd
--
-- Remarks
--
return ci_plsql_modules.id%type
is
   l_plm_id              ci_plsql_modules.id%type;
   l_plm_already_exists  boolean;
   l_purpose             ci_plsql_modules.purpose%type;
   l_validation_code     varchar2(4000);
   l_arguments           varchar2(2000); -- 6.5.3.3AMIS
begin
   bllog.write('starting maintain_plm', bllog.debug);
   if p_validation_code is null -- 6.5.3.2
   then
      l_validation_code := get_default_code;
   else
      l_validation_code := p_validation_code;
   end if;

   l_purpose := substr
                ( p_description
                , 1
                , MAX_LENGTH_PURPOSE
                );

   -- check if a PL/SQL Definition with this short name already exists
   l_plm_already_exists := hsu_capi.find_plm           -- 6.5.0.1  extra parameters added
                           ( p_plm_short_name => p_br_name            -- in
                           , p_tbl_id         => p_tbl_id             -- in
                           , p_type           => 'BRDD'               -- in
                           , p_plm_id         => l_plm_id             -- out
                           );

   if not l_plm_already_exists
   then
      -- a new PL/SQL Definition must be created
      l_plm_id := blplm.ins
                  ( short_name                  => p_br_name
                  , name                        => p_br_name
                  , purpose                     => l_purpose
                  , plsql_module_type           => 'TRG-LOGIC'
                  , implementation_name         => lower(p_br_name)
                  , free_format_text_flag       => 'Y'
                  );
      bllog.write
      ( 'Created PL/SQL Definition '
        ||hsu_name.get_name_and_path(l_plm_id) -- 6.5.0.0  path added by special function
        ||'.'
      , bllog.information
      );
   else
      -- recreate the existing PL/SQL Definition
      -- update everything (except short name, because that's how it was identified)
      blplm.upd
      ( p_plm_id              => l_plm_id
      , name                  => p_br_name
      , purpose               => l_purpose
      , plsql_module_type     => 'TRG-LOGIC'
      , implementation_name   => lower(p_br_name)
      , free_format_text_flag => 'Y'
      );
      bllog.write
      ( 'Recreated existing PL/SQL Definition '
        ||hsu_name.get_name_and_path(l_plm_id)   -- 6.5.0.0  path added
        ||', preserving only the Revision History.'
      , bllog.information
      );
      for r_arg in blarg.c_arg_plm(l_plm_id) loop
         blarg.del(r_arg.arg_id);
      end loop;
      -- delete existing arguments
      bllog.write('Existing arguments deleted.', bllog.debug);
   end if; -- l_plm_already_exists


   add_notes(l_plm_id, p_rule_type, p_message_code);

   add_arguments(l_plm_id, p_col_data_required, 'p_');
   l_arguments:= substr(get_arguments,2); -- 6.5.3.3AMIS

   add_arguments(l_plm_id, p_old_col_data_required, 'p_old_');
   l_arguments:= l_arguments||substr(get_arguments,2); -- 6.5.3.3AMIS

   add_plsql_block
   ( l_plm_id
   , l_plm_already_exists
   , p_description
   , p_utility_package
   , p_utility_revision
   , p_validation_code
   , p_br_name
   , l_arguments -- 6.5.3.3AMIS
   );

   bllog.write('end of maintain_plm', bllog.debug);
   return l_plm_id;
end maintain_plm;


function maintain_trg
( p_br_name            in varchar2
, p_tbl_id             in ci_table_definitions.id%type
, p_action             in varchar2
, p_triggering_columns in varchar2
, p_when_condition     in varchar2
, p_plm_id             in ci_plsql_modules.id%type
, p_description        in varchar2
)
-- Purpose  Maintains Database Trigger part of Business Rule Design Definition
--          and returns its id
--
-- Usage    Called from maintain_brdd
--
-- Remarks
--
return ci_database_triggers.id%type
is
   l_purpose             ci_database_triggers.purpose%type;
   r_tbl                 bltbl.c_tbl%rowtype;
   r_col                 blcol.c_col%rowtype;
   r_trgcol              bltrgcol.c_trgcol_ob_trg%rowtype;
   l_trgcol_id           ci_trigger_column_usages.id%type;
   l_trg_id              ci_database_triggers.id%type;
   l_trg_already_exists  boolean;
   l_trigger_on_delete   ci_database_triggers.trigger_on_delete%type;
   l_trigger_on_insert   ci_database_triggers.trigger_on_insert%type;
   l_trigger_on_update   ci_database_triggers.trigger_on_update%type;
   l_col_table           hsu_par.list_items_table_type;
begin
   -- retrieve table properties
   open bltbl.c_tbl(p_tbl_id);
   fetch bltbl.c_tbl into r_tbl;
   close bltbl.c_tbl;

   -- determine purpose
   l_purpose := substr
                ( p_description
                , 1
                , MAX_LENGTH_PURPOSE
                );

   -- determine trigger_on_delete/insert/update
   if instr(p_action, 'D') > 0
   then
      l_trigger_on_delete := 'Y';
   else
      l_trigger_on_delete := 'N';
   end if;
   if instr(p_action, 'I') > 0
   then
      l_trigger_on_insert := 'Y';
   else
      l_trigger_on_insert := 'N';
   end if;
   if instr(p_action, 'U') > 0
   then
      l_trigger_on_update := 'Y';
   else
      l_trigger_on_update := 'N';
   end if;

   -- check if trigger with this name already exists
   -- (and check if it belongs to the right table)
   l_trg_already_exists := hsu_capi.find_trg
                           ( p_br_name
                           , p_tbl_id
                           , r_tbl.tbl_name
                           , l_trg_id
                           );

   if not l_trg_already_exists
   then
      -- insert trigger
      l_trg_id := bltrg.ins
                  ( table_definition_reference => p_tbl_id
                  , plsql_module_reference     => p_plm_id
                  , complete_flag              => 'N'
                  , execute_time               => 'AFTER' -- documentation only
                  , name                       => p_br_name
                  , purpose                    => l_purpose
                  , trigger_level              => 'STATEMENT' -- documentation only
                  , trigger_on_delete          => l_trigger_on_delete
                  , trigger_on_insert          => l_trigger_on_insert
                  , trigger_on_update          => l_trigger_on_update
                  );
      bllog.write
      ( 'Created trigger '
        ||upper(p_br_name)||' for table '
        ||hsu_name.get_name_and_path(r_tbl.tbl_irid) -- 6.5.0.0  path added
        ||'.'
      , bllog.information
      );
   else
      -- trigger exists, so update it
      bltrg.upd
      ( p_trg_id                   => l_trg_id
      , plsql_module_reference     => p_plm_id
      , complete_flag              => 'N'
      , execute_time               => 'AFTER' -- documentation only
      , purpose                    => l_purpose
      , trigger_level              => 'STATEMENT' -- documentation only
      , trigger_on_delete          => l_trigger_on_delete
      , trigger_on_insert          => l_trigger_on_insert
      , trigger_on_update          => l_trigger_on_update
      );
      bllog.write
      ( 'Recreated existing trigger '||upper(p_br_name)||' for table '
        ||hsu_name.get_name_and_path(r_tbl.tbl_irid) -- 6.5.0.0  path added
        ||'.'
      , bllog.information
      );
      -- delete existing trigger column usages
      <<trgcol>>
      for r_trgcol in bltrgcol.c_trgcol_ob_trg(l_trg_id) loop
         bltrgcol.del(r_trgcol.trgcol_id);
      end loop trgcol;
   end if; -- l_trg_already_exists

   -- add when condition
   bltext.init_text_lines(g_text, g_text_counter);
   add(p_when_condition);
   bltext.write_lines_to_repository
   ( l_trg_id
   , 'TRG'
   , 'CDIWHN'
   , g_text
   , g_text_counter
   , 'REPLACE'
   );

   -- add trigger column usages
   if l_trigger_on_update = 'Y'
   then
      l_col_table := hsu_par.get_list_items(p_triggering_columns, '|');
      <<col>>
      for i in 1 .. nvl(l_col_table.last, 0) loop
         open blcol.c_col(to_number(l_col_table(i)));
         fetch blcol.c_col into r_col;
         close blcol.c_col;
         l_trgcol_id := bltrgcol.ins
                        ( database_trigger_reference => l_trg_id
                        , column_reference           => r_col.col_id
                        );
         bllog.write
         ( 'Trigger column usage added for column '||r_col.col_name||'.'
         , bllog.information
         );
      end loop col;
   end if; -- l_trigger_on_update

   return l_trg_id;
exception
   when others
   then
      raise;
end maintain_trg;


function create_message_if_needed
-- Purpose  If message number is given, create message with that number
--          if it does not exist yet.
--          If message number is not given, find first free number within prefix
--          and create that message.
--          Returns the resulting message code
--
-- Usage    from maintain_brdd
--
-- Remarks  When creating or looking for message, prefix br name with app code
--          because constraint name in message should be unique over all applications
--
( p_message_prefix    in varchar2
, p_message_number_c  in varchar2
, p_message_language  in varchar2
, p_description       in varchar2
, p_rule_type         in varchar2 -- 6.5.3.3  new parameter
, p_constraint_name   in varchar2 -- 6.5.3.3  new parameter
)
return varchar2
is
   l_message_code      varchar2(20);
   l_existing_msg_code varchar2(20);
   l_create_message    boolean      := false;
   l_description       varchar2(2000);
   l_message_text      varchar2(2000);
   l_constraint_name   varchar2(100);  -- 6.5.3.3 new variable
begin
   if p_message_number_c is not null
   then
      l_message_code := upper(p_message_prefix)||'-'||p_message_number_c;
      -- check if message code does not exist in message table
      if hil_message.msg_code_exists
      ( p_msg_code          => l_message_code
      , p_existing_msg_code => l_existing_msg_code
      )
      then
         bllog.write
         ( 'Message '||l_message_code
           ||' already exists in Headstart message tables, no new message was created.'
         , bllog.information
         );
      else
         l_create_message := true;
      end if; -- msg_code exists
   elsif hil_message.msg_code_exists
         ( p_constraint        => p_constraint_name  -- 6.5.3.3 changed
         , p_existing_msg_code => l_existing_msg_code
         )
   then
      l_message_code := l_existing_msg_code;
      -- no message number, but message for this business rule already exists
      bllog.write
      ( 'Existing Message '||l_existing_msg_code
        ||' was already linked to constraint name '
        ||p_constraint_name||', so that message is used.' -- 6.5.3.3 changed
      , bllog.information
      );
   else
      -- find free number
      l_message_code := hil_message.get_next_msg_code(p_message_prefix);
      bllog.write('Next free message code: '||l_message_code, bllog.debug_detailed);
      l_create_message := true;
   end if; -- p_message_number
   if l_create_message
   then
      -- 6.5.0.0  final check on the message format
      if not qms$errors.CheckErrorFormat
             ( p_msgtxt => l_message_code
             )
      then
         -- message code is not correct
         bllog.write
         ( 'The message code '''||l_message_code||''' is not correct:'
           ||chr(10)
           ||FORMAT_RATIONALE_1
         , bllog.errors
         );
         bllog.write -- split up in 3 pieces (because of max length of text in bllog)
         ( FORMAT_RATIONALE_2
         , bllog.errors
         );
         bllog.write
         ( FORMAT_RATIONALE_3
         , bllog.errors
         );
         -- abort utility
         raise e_abort_utility;
      end if;
      l_description := substr
                       ( replace(p_description, chr(10), ' ')
                       , 1
                       , MAX_LENGTH_MSG_DESCR
                       );
      -- 6.5.3.3 divert based on p_rule_type
      if p_rule_type = 'C' -- check constraint
      then
         l_message_text := substr
                           ( l_description
                           , 1
                           , MAX_LENGTH_MSG_TEXT
                           );
      else
         l_message_text := substr
                           ( '<p1>: '||l_description
                           , 1
                           , MAX_LENGTH_MSG_TEXT
                           );
      end if;
      -- end change 6.5.3.3

      bllog.write
      ( 'Creating message '
        ||l_message_code
        ||' in Headstart Message tables with text '''
        ||htf.escape_sc(l_message_text)
        ||''''
      , bllog.information
      );
      hil_message.create_message
      ( p_code            => l_message_code
      , p_description     => l_description
      , p_language        => p_message_language
      , p_text            => l_message_text
      , p_constraint_name => p_constraint_name -- 6.5.3.3 changed
      );
   end if; -- l_create_message

   return l_message_code;
end create_message_if_needed;


--
-- public procedures and functions
--


procedure deinstall
-- Purpose  removes the package from the Headstart Utilities tables.
--
-- Usage
--
-- Remarks
--
is
begin
   hsu_install.deinstall(PACKAGE_NAME);
end deinstall;


procedure install
-- Purpose  installs the package into the Headstart Utilities tables
--
-- Usage
--
-- Remarks
--
is
begin
   hsu_install.add_utility
   ( PACKAGE_NAME
   , UTILITY_NAME
   , 'PB'
   ,  -- help text
     'This Productivity Booster is part of a suite of utilities in support of CDM '
   ||'RuleFrame.  See the CDM Standards and Guidelines Library'
   ||' and the Headstart User Guide for more information.'
   ||chr(10)
   ||chr(10)
   ||'This utility allows you to create'
   ||' a Business Rule Design Definition, consisting of a Database Trigger'
   ||' and a corresponding PL/SQL Definition.'
   ||chr(10)
   ||' If the Business Rule Design Definition already exists in'
   ||' the context application system, it is recreated,'
   ||' overwriting all components of the existing definition.'
   ||' The only part that is preserved is the Revision History'
   ||' in the PL/SQL Block of the PL/SQL Definition, which is identified as the text'
   ||' between the phrase ''Revision History'' and the end of the comment block'
   ||' (''***********************************************/'').'
   ||' A new record will be added to the Revision History just below the heading,'
   ||' which is assumed to be the first 4 lines of the Revision History.'
   ||chr(10)
   ||chr(10)
   ||'If you want to modify an existing Business Rule Design Definition'
   ||' (instead of recreating it),'
   ||' modify the relevant Database Trigger and/or PL/SQL Definition directly.'
   ||chr(10)
   ||chr(10)
   ||' The Business Rule Design Definitions are used as input for the'
   ||' utility that generates the Custom API (CAPI).'
   ||' The only exception to this is when you specify the Rule Type to be'
   ||' "Check Constraint", then a database check constraint is created.'
   );

   hsu_install.add_parameter
   ( PACKAGE_NAME                 -- package name
   , 1                            -- sequence
   , 'Business Rule Name'         -- prompt
   , null                         -- default value actual
   , null                         -- default value displayed
   , 'Y'                          -- mandatory
   , 'N'                          -- allow multi-select
   , 'Y'                          -- include shared elements
   , null                         -- element type short name
   , null                         -- sql expression actual
   , null                         -- sql expression displayed
   , null                         -- where clause
   , null                         -- synchronize with
   , null                         -- foreign key column
   , 'Name of the Business Rule (e.g. BR_EMP001_TPL). Maximum length: '
     ||to_char(MAX_LENGTH_BR_NAME)
     ||'.'
                                  -- help text
   );

   hsu_install.add_parameter
   ( PACKAGE_NAME                 -- package name
   , 2                            -- sequence
   , 'Description'                -- prompt
   , null                         -- default value actual
   , null                         -- default value displayed
   , 'Y'                          -- mandatory
   , 'N'                          -- allow multi-select
   , 'N'                          -- include shared elements
   , ''                           -- element type short name
   , ''                           -- sql expression actual
   , ''                           -- sql expression displayed
   , null                         -- where clause
   , null                         -- synchronize with
   , null                         -- foreign key column
   , 'Description of the Business Rule that will be written as Purpose'
     ||' into the code comments in the PL/SQL Block.'
     ||' The first '
     ||MAX_LENGTH_PURPOSE
     ||' characters are also written to the Purpose property of the Database Trigger'
     ||' and the corresponding PL/SQL Definition,'
     ||' or the Comment of the Check Constraint.'
     ||chr(10)
     ||' If the message code did not exist yet, the first '
     ||MAX_LENGTH_MSG_TEXT
     ||' characters are also used for the error message.'
                                  -- help text
   );

   hsu_install.add_parameter
   ( PACKAGE_NAME                 -- package name
   , 3                            -- sequence
   , 'Table'                      -- prompt
   , null                         -- default value actual
   , '<null>'                     -- default value displayed
   , 'Y'                          -- mandatory
   , 'N'                          -- allow multi-select
   , 'Y'                          -- include shared elements
   , 'TBL'                        -- element type short name
   , 'TO_CHAR(ID)'                -- sql expression actual
   , 'NAME'                       -- sql expression displayed
   , null                         -- where clause
   , null                         -- synchronize with
   , null                         -- foreign key column
   , 'Choose the table in whose CAPI the Business Rule should be added,'
   ||' or to which the Check Constraint should be added.'
                                  -- help text
   );

   hsu_install.add_parameter
   ( PACKAGE_NAME                 -- package name
   , 4                            -- sequence
   , 'Type of Rule'               -- prompt
   , 'S'                          -- default value actual
   , 'Static Data Constraint'     -- default value displayed
   , 'Y'                          -- mandatory
   , 'N'                          -- allow multi-select
   , 'N'                          -- include shared elements
   , ''                           -- element type short name
   , ''                           -- sql expression actual
   , ''                           -- sql expression displayed
   , null                         -- where clause
   , null                         -- synchronize with
   , null                         -- foreign key column
   ,   'Indicate the classification of the Business Rule.'
     ||chr(10)||' '
     ||chr(10)||'Data Constraint rules enforce the validity of data.  A static rule is '
     ||'always true and can be checked via a ''validate all'' service.  A dynamic '
     ||'rule is true only at the moment the operation is performed.'
     ||'  A rule that can not be rewritten without a reference to any of the following'
     ||' is always dynamic:'
     ||chr(10)||'  operation (insert, update or delete)'
     ||chr(10)||'  old column values'
     ||chr(10)||'  user'
     ||chr(10)||'  sysdate'
     ||chr(10)||'Change Event rules trigger changes to the data or system.'
     ||'  A Change Event '
     ||'with DML involves an insert, update or delete to the database.  This may '
     ||'involve derivations, denormalizations, journaling, etc.  A Change Event '
     ||'without DML involves trigger an action external to the database such as '
     ||'submitting a batch job or sending an email.'
     ||chr(10)||chr(10)
     ||'Choose Check Constraint if you want to implement the rule outside the CAPI'
     ||' in a server side check constraint (better performance).'
     ||' This is an alternative for certain'
     ||' Attribute or Tuple rules (which are both Static rules).'
                                  -- help text
   );

   hsu_install.add_allowable_value
   ( PACKAGE_NAME                 -- package name
   , 4                            -- parameter
   , 1                            -- sequence
   , 'S'                          -- actual value
   , 'Static Data Constraint'     -- displayed value
   );

   hsu_install.add_allowable_value
   ( PACKAGE_NAME                 -- package name
   , 4                            -- parameter
   , 2                            -- sequence
   , 'D'                          -- actual value
   , 'Dynamic Data Constraint'    -- displayed value
   );

   hsu_install.add_allowable_value
   ( PACKAGE_NAME                 -- package name
   , 4                            -- parameter
   , 3                            -- sequence
   , 'W'                          -- actual value
   , 'Change Event with DML'      -- displayed value
   );

   hsu_install.add_allowable_value
   ( PACKAGE_NAME                 -- package name
   , 4                            -- parameter
   , 4                            -- sequence
   , 'O'                          -- actual value
   , 'Change Event without DML'   -- displayed value
   );

   hsu_install.add_allowable_value
   ( PACKAGE_NAME                 -- package name
   , 4                            -- parameter
   , 5                            -- sequence
   , 'C'                          -- actual value
   , 'Check Constraint (non-CAPI)'-- displayed value
   );

   hsu_install.add_parameter
   ( PACKAGE_NAME                 -- package name
   , 5                            -- sequence
   , 'Triggering DML Event'      -- prompt
   , 'I|'                         -- default value actual
   , 'Insert'                     -- default value displayed
   , 'Y'                          -- mandatory
   , 'Y'                          -- allow multi-select
   , 'N'                          -- include shared elements
   , ''                           -- element type short name
   , ''                           -- sql expression actual
   , ''                           -- sql expression displayed
   , null                         -- where clause
   , null                         -- synchronize with
   , null                         -- foreign key column
   ,   'Indicate which DML event(s) will trigger enforcement of the Business '
     ||'Rule.'
                                  -- help text
   );
   hsu_install.add_allowable_value
   ( PACKAGE_NAME                 -- package name
   , 5                            -- parameter
   , 1                            -- sequence
   , 'I'                          -- actual value
   , 'Insert'                     -- displayed value
   );
   hsu_install.add_allowable_value
   ( PACKAGE_NAME                 -- package name
   , 5                            -- parameter
   , 2                            -- sequence
   , 'U'                          -- actual value
   , 'Update'                     -- displayed value
   );
   hsu_install.add_allowable_value
   ( PACKAGE_NAME                 -- package name
   , 5                            -- parameter
   , 3                            -- sequence
   , 'D'                          -- actual value
   , 'Delete'                     -- displayed value
   );

   hsu_install.add_parameter
   ( PACKAGE_NAME                 -- package name
   , 6                            -- sequence
   , 'Update Trigger Columns'    -- prompt
   , null                         -- default value actual
   , '<null>'                     -- default value displayed
   , 'N'                          -- mandatory
   , 'Y'                          -- allow multi-select
   , 'N'                          -- include shared elements
   , 'COL'                        -- element type short name
   , 'TO_CHAR(ID)'                -- sql expression actual
   , 'NAME'                       -- sql expression displayed
   , ''                           -- where clause
   , 3                            -- synchronize with
   , 'TABLE_REFERENCE'            -- foreign key column
   ,   'Choose the columns whose change of value will trigger enforcement of the '
     ||'Business Rule (only relevant if Update is a triggering event).'
                                  -- help text
   );

   hsu_install.add_parameter
   ( PACKAGE_NAME                 -- package name
   , 7                            -- sequence
   , 'When Condition'             -- prompt
   , null                         -- default value actual
   , null                         -- default value displayed
   , 'N'                          -- mandatory
   , 'N'                          -- allow multi-select
   , 'N'                          -- include shared elements
   , ''                           -- element type short name
   , ''                           -- sql expression actual
   , ''                           -- sql expression displayed
   , null                         -- where clause
   , null                         -- synchronize with
   , null                         -- foreign key column
   , 'Additional conditions specifying when the business rule needs to be enforced.'
     ||' The conditions entered here will be checked at the moment of triggering,'
     ||' as opposed to the moment the transaction closes.'
     ||chr(10)
     ||chr(10)||'It must be a boolean expression, and it may contain the following'
     ||' constructs:'
     ||chr(10)||' <I>:new.[column name]</I> for columns existing in the table,'
     ||' indicating the (new) value of the column'
     ||chr(10)||' <I>:old.[column name]</I> for columns existing in the table,'
     ||' indicating the old value of the column'
     ||chr(10)||' any of the booleans <I>inserting</I>, <I>updating</I>'
     ||' or <I>deleting</I>'
     ||chr(10)||' the pseudo columns <I>sysdate</I> and <I>user</I>'
     ||chr(10)||' calls to function <I>qms_transaction_mgt.not_on_stack_yet</I>'
     ||chr(10)||'Note: In this expression, you should not access any table data'
     ||' except by the :new.[column name] or :old.[column name] construction,'
     ||' nor should any procedures/functions be called that access table data.'
                                  -- help text
   );

   hsu_install.add_parameter
   ( PACKAGE_NAME                 -- package name
   , 8                            -- sequence
   , 'Column Args for Validation' -- prompt
   , null                         -- default value actual
   , '<null>'                     -- default value displayed
   , 'Y'                          -- mandatory
   , 'Y'                          -- allow multi-select
   , 'Y'                          -- include shared elements
   , 'COL'                        -- element type short name
   , 'TO_CHAR(ID)'                -- sql expression actual
   , 'NAME'                       -- sql expression displayed
   , ''                           -- where clause
   , 3                            -- synchronize with
   , 'table_reference'            -- foreign key column
   , 'Choose the columns whose new values need to be supplied as arguments (parameters)'
     ||' for the validation of the Business Rule.'
     ||chr(10)||chr(10)
     ||'If the Rule Type is Check Constraint, this parameter is ignored.'
                                  -- help text
   );

   hsu_install.add_parameter
   ( PACKAGE_NAME                 -- package name
   , 9                            -- sequence
   , 'Old Column Args for Validation' -- prompt
   , null                         -- default value actual
   , '<null>'                     -- default value displayed
   , 'N'                          -- mandatory
   , 'Y'                          -- allow multi-select
   , 'Y'                          -- include shared elements
   , 'COL'                        -- element type short name
   , 'TO_CHAR(ID)'                -- sql expression actual
   , 'NAME'                       -- sql expression displayed
   , ''                           -- where clause
   , 3                            -- synchronize with
   , 'table_reference'            -- foreign key column
   , 'Choose the columns whose <B>old</B> or pre(vious) values need to be supplied'
     ||' as arguments (parameters)'
     ||' for the validation of the Business Rule.'
                                  -- help text
   );

   hsu_install.add_parameter
   ( PACKAGE_NAME                 -- package name
   , 10                           -- sequence
   , 'Validation Code'            -- prompt
   , get_default_code             -- default value actual
   , get_default_code             -- default value displayed
   , 'Y'                          -- mandatory
   , 'N'                          -- allow multi-select
   , 'N'                          -- include shared elements
   , ''                           -- element type short name
   , ''                           -- sql expression actual
   , ''                           -- sql expression displayed
   , null                         -- where clause
   , null                         -- synchronize with
   , null                         -- foreign key column
   , 'PL/SQL code performing the actual enforcement of the Business Rule.'
     ||' The code entered here will be checked at the moment the transaction closes,'
     ||' as opposed to the moment of triggering.'
     ||chr(10)
     ||chr(10)
     ||'Enter code that sets l_rule_ok to false when the Business Rule is violated.'
     ||' You can make use of p_[column name] and p_old_[column name]'
     ||' for the (old) columns that were specified as (Old) Column Args for Validation.'
     ||' For example, if the rule is: <I>Shippings are not allowed to be delayed</I>,'
     ||' the validation code could contain:'
     ||' <I>l_rule_ok := p_date_shipped <= p_old_date_shipped;</I>'
     ||chr(10)
     ||chr(10)
     ||'If the Rule Type is Check Constraint, enter here the Where/Validation Condition'
     ||' for the Check Constraint.'
     ||chr(10)
     ||chr(10)
     ||' The utility will replace [br_name] with the Business Rule name'
     ||' as entered in the first parameter.'
                                       -- help text
   );

   hsu_install.add_parameter
   ( PACKAGE_NAME                 -- package name
   , 11                           -- sequence
   , 'Message Prefix'             -- prompt
   , '<default>'                  -- default value actual
   , '<default>'                  -- default value displayed -- 6.5.0.0  changed to <default>
   , 'Y'                          -- mandatory
   , 'N'                          -- allow multi-select
   , 'N'                          -- include shared elements
   , ''                           -- element type short name
   , ''                           -- sql expression actual
   , ''                           -- sql expression displayed
   , null                         -- where clause
   , null                         -- synchronize with
   , null                         -- foreign key column
   ,  'Specify the prefix for the code of the message that should be displayed upon '
    ||'violation of this business rule. If the value is '''||htf.escape_sc('<default>')
    ||''', the utility will take the three letter abbreviation of the owning container '
    ||'where the pl/sql module for the design definition is created.'
                                  -- help text
   );

   hsu_install.add_parameter
   ( PACKAGE_NAME                 -- package name
   , 12                           -- sequence
   , 'Message Number'             -- prompt
   , null                         -- default value actual
   , null                         -- default value displayed
   , 'N'                          -- mandatory
   , 'N'                          -- allow multi-select
   , 'N'                          -- include shared elements
   , ''                           -- element type short name
   , ''                           -- sql expression actual
   , ''                           -- sql expression displayed
   , null                         -- where clause
   , null                         -- synchronize with
   , null                         -- foreign key column
   ,  'If known, specify the number that together with the message prefix forms a'
    ||' message code, to identify the message that should be displayed upon '
    ||'violation of this business rule. If a message with this code is not'
    ||' present in the Headstart Message tables, it will be created.'
    ||chr(10)
    ||chr(10)
    ||' If no message number is specified, the next free number within the message prefix'
    ||' is used to create a new message.'
                                  -- help text
   );

   hsu_install.add_msg_language_parameter
   ( PACKAGE_NAME                 -- package name
   , 13                           -- sequence
   , 'Message Language'           -- prompt
   , -- help text
     'Specify the language that the message must be created in'
     ||' if the message code does not exist yet in the Headstart message tables.'
   );
end install;



function revision
return varchar2
-- Purpose  returns the revision label of the package
--
-- Usage
--
-- Remarks
--
is
begin
   return blvers.get_revision_label(REVISION_LABEL);
end revision;


function get_default_code
return varchar2
is
begin
   return DEFAULT_CODE;
end get_default_code;


procedure maintain_brdd
( p_utility_package       in  varchar2
, p_utility_revision      in  varchar2
, p_br_name               in  varchar2
, p_description           in  varchar2
, p_tbl_id                in  number
, p_tbl_alias             in  varchar2 default null        -- 6.5.3.2 new
, p_rule_type             in  varchar2 default 'S'
, p_action                in  varchar2 default 'I|'
, p_triggering_columns    in  varchar2 default null
, p_when_condition        in  varchar2 default null
, p_col_data_required     in  varchar2
, p_old_col_data_required in  varchar2 default null
, p_message_prefix        in  varchar2 default bl.NO_VARCHAR2_VALUE_PASSED -- fix for Oracle 9i
, p_message_number        in  varchar2 default null
, p_message_language      in  varchar2 default 'ENG'
, p_validation_code       in  varchar2 default null        -- 6.5.3.2 changed default
, p_trg_id                out ci_database_triggers.id%type
, p_plm_id                out ci_plsql_modules.id%type
, p_cco_id                out ci_check_constraints.id%type -- 6.5.3.2 new
)
-- Purpose  Maintains Business Rule Design Definition and returns id's of relevant objects
--
-- Usage    Called from hsu_crbr.run and hsu_brtr.run
--          p_br_name is the name of the business rule being created/updated
--          p_description is the description of the business rule
--          p_tbl_id is the id of the table whose CAPI has to contain the business rule
--          p_rule_type is S(tatic Data Constraint), D(ynamic Data Constraint)
--             , W (Change Event with DML), or O (Change Event without DML)
-- 6.5.3.2     or C (Check Constraint)
--          p_action is a string with any combination of I(nsert),U(pdate) and D(elete)
--          p_trigger_columns is a list of column ids that when updated trigger business
--             rule enforcement; only relevant when p_action contains 'U'
--          p_when_condition describes additional conditions for triggering event
--             next to action and changes in triggering columns
--          p_col_data_required is a list of column ids whose values are required for
--             validation of this Business Rule
--          p_old_col_data_required is a list of column ids whose values are required for
--             validation of this Business Rule
--          p_message_prefix is the prefix of the message in the Headstart Message tables
--             that should be given when the rule is violated
--          p_message_number is the suffix of the message in the Headstart Message tables
--             that should be given when the rule is violated
--          p_message_language is the language that should be used for creation of the
--             message if the message number is not given or didn't exist yet
--          p_validation_code is the PL/SQL code to validate the rule,
--             where you may assume that the variable l_rule_ok is already declared
--             and will be the return value of the validation function
--          p_trg_id returns the id of the Database Trigger representing the Bus.Rule (if any)
--          p_plm_id returns the id of the PL/SQL Definition representing the Bus.Rule (if any)
-- 6.5.3.2  p_cco_id returns the id of the Check Constraint representing the Bus.Rule (if any)
--
-- Remarks  p_action, p_trigger_columns, p_col_data_required and p_old_col_data_required
--          are multi-select parameters, which means that the individual values must be
--          concatenated, separated by and ending with '|' characters.
--          Example: '34|42|683|'.
--
is
   l_message_code     varchar2(15);
   l_message_prefix   varchar2(30);
   l_context_app_code varchar2(30);
   l_validation_code  varchar2(2000);
   l_constraint_name  varchar2(100); -- 6.5.3.3 new
begin
   -- prefix this message with html newline to separate new brdd from previous
   bllog.write('<BR>Business Rule Design Definition for '||upper(p_br_name));

   -- 6.5.0.0  if default context folder, set it to folder of table
   if blapi.get_default_context_fol_flag
   then
      blapi.set_context_fol
      ( p_context_folder_irid => jr_acc_rights.get_owning_container(p_tbl_id)
      );
   end if;

   --
   -- 6.5.0.0  get application system code for context container
   l_context_app_code := hsu_name.get_app_code
                         ( jr_name.get_crn_from_irid
                           ( blapi.get_context_fol_irid
                           , 'NAME'
                           )
                         );
   -- use appropriate message code
   if p_message_prefix = '<default>'
   then
      l_message_prefix := l_context_app_code;
      bllog.write
      ( 'Message prefix set to default '||l_message_prefix
      , bllog.debug
      );
   else
      l_message_prefix := p_message_prefix;
   end if;

   initial_checks
   ( p_br_name
   , p_rule_type
   , p_action
   , p_triggering_columns
   , p_col_data_required
   , p_old_col_data_required
   , l_message_prefix -- 6.5.0.0  changed
   , p_message_number
   , l_context_app_code -- 6.5.1.0  added
   );

   -- 6.5.3.3 determine message constraint name and pass it to create_message...
   --         for brdd convention is: <APP alias>_<BR name>
   --         for cco  convention is: <APP alias>_<Table alias>_<BR name>, which
   --                                 exactly equals the check constraint name
   if p_rule_type = 'C'
   then
      -- check constraint
      l_constraint_name := hsu_name.get_next_cco_name
                           ( p_table_id      => p_tbl_id
                           , p_table_alias   => p_tbl_alias
                           , p_logical_name  => p_br_name
                           );
   else
      -- br design definition
      l_constraint_name := upper(l_context_app_code||'_'||p_br_name);
   end if;
   bllog.write
   ( 'Create/maintain message with constraint_name = '||l_constraint_name
   , bllog.debug
   );
   l_message_code := create_message_if_needed
                     ( l_message_prefix -- 6.5.0.0  changed
                     , p_message_number
                     , p_message_language
                     -- , p_br_name                -- 6.5.3.3 removed
                     , p_description
                     -- , l_context_app_code       -- 6.5.3.3 removed
                     , p_rule_type
                     , l_constraint_name           -- 6.5.3.3 new
                     );

   -- 6.5.3.2 Possibility to create rule as check constraint
   if p_rule_type = 'C' -- check constraint
   then
      p_cco_id := maintain_cco
                  ( p_context_app_code       => l_context_app_code
                  , p_br_name                => p_br_name
                  , p_tbl_id                 => p_tbl_id
                  , p_tbl_alias              => p_tbl_alias
                  , p_description            => p_description
                  , p_message_code           => l_message_code
                  , p_when_condition         => p_when_condition
                  , p_validation_code        => p_validation_code
                  );
   else -- don't create check constraint, create PLM and TRG
      -- 6.5.3.3 for p_rule_type <> 'C' and empty validation code, set to default
      if p_validation_code is null
      then
         l_validation_code := get_default_code;
      else
         l_validation_code := p_validation_code;
      end if;

      p_plm_id := maintain_plm
                  ( p_br_name
                  , p_tbl_id             -- 6.5.0.1  parameter added
                  , p_description
                  , p_rule_type
                  , p_col_data_required
                  , p_old_col_data_required
                  , l_message_code
                  , p_utility_package
                  , p_utility_revision
                  , l_validation_code    -- 6.5.3.3. changed
                  );
      p_trg_id := maintain_trg
                  ( p_br_name
                  , p_tbl_id
                  , p_action
                  , p_triggering_columns
                  , p_when_condition
                  , p_plm_id
                  , p_description
                  );
   end if; -- p_rule_type

   bllog.write
   ('Maintain_brdd output parameter p_trg_id = '||to_char(p_trg_id)
    ||', p_plm_id = '||to_char(p_plm_id)
    ||', p_cco_id = '||to_char(p_cco_id)
   , bllog.debug
   );
exception
   when others then
      raise;
end maintain_brdd;

procedure run
( p_br_name               in varchar2
, p_description           in varchar2
, p_tbl_id                in number
, p_rule_type             in varchar2 default 'S'
, p_action                in varchar2 default 'I|'
, p_triggering_columns    in varchar2 default null
, p_when_condition        in varchar2 default null
, p_col_data_required     in varchar2
, p_old_col_data_required in varchar2 default null
, p_validation_code       in varchar2 default get_default_code
, p_message_prefix        in varchar2 default '<default>'
, p_message_number        in varchar2 default null
, p_message_language      in varchar2 default 'ENG'
)
-- Purpose  Runs the utility
--
-- Usage    For usage of parameters, see procedure maintain_brdd
--
is
   l_dummy_trg_id ci_database_triggers.id%type;
   l_dummy_plm_id ci_plsql_modules.id%type;
   l_dummy_cco_id ci_check_constraints.id%type;
begin
   bllog.write('Start of hsu_crbr.run', bllog.debug);
   -- call the procedure that does the work,
   -- in this case we don't need the out parameters so we provide dummy variables
   maintain_brdd
   ( PACKAGE_NAME
   , revision
   , p_br_name
   , p_description
   , p_tbl_id
   , null -- p_tbl_alias
   , p_rule_type
   , p_action
   , p_triggering_columns
   , p_when_condition
   , p_col_data_required
   , p_old_col_data_required
   , p_message_prefix
   , p_message_number
   , p_message_language
   , p_validation_code
   , l_dummy_trg_id
   , l_dummy_plm_id
   , l_dummy_cco_id
   );
   bllog.write('Completed hsu_crbr.run', bllog.debug);
exception
   when others then
      if blapi.handle_exception then
         raise;
      end if;
end run;

end hsu_crbr;
